"""
Interfaces to various optimizers.
"""

import sys
import warnings

# CRUFT: time.clock() removed from python 3.8
try:
    from time import perf_counter
except ImportError:
    from time import clock as perf_counter

import numpy as np

from . import monitor
from . import initpop
from . import lsqerror

from .history import History
from .formatnum import format_uncertainty
from .fitproblem import nllf_scale

from .dream import MCMCModel


class ConsoleMonitor(monitor.TimedUpdate):
    """
    Display fit progress on the console
    """

    def __init__(self, problem, progress=1, improvement=30):
        monitor.TimedUpdate.__init__(self, progress=progress, improvement=improvement)
        self.problem = problem

    def show_progress(self, history):
        scale, err = nllf_scale(self.problem)
        chisq = format_uncertainty(scale * history.value[0], err)
        print("step", history.step[0], "cost", chisq)
        sys.stdout.flush()

    def show_improvement(self, history):
        # print("step",history.step[0],"chisq",history.value[0])
        p = self.problem.getp()
        try:
            self.problem.setp(history.point[0])
            print(self.problem.summarize())
        finally:
            self.problem.setp(p)
        sys.stdout.flush()


class CheckpointMonitor(monitor.TimedUpdate):
    """
    Periodically save fit state so that it can be resumed later.
    """

    #: Function to call at each checkpoint.
    checkpoint = None  # type: Callable[None, None]

    def __init__(self, checkpoint, progress=60 * 30):
        monitor.TimedUpdate.__init__(self, progress=progress, improvement=np.inf)
        self.checkpoint = checkpoint
        self._first = True

    def show_progress(self, history):
        # Skip the first checkpoint since it only contains the
        # start/resume state
        if self._first:
            self._first = False
        else:
            self.checkpoint(history)

    def show_improvement(self, history):
        pass


class StepMonitor(monitor.Monitor):
    """
    Collect information at every step of the fit and save it to a file.

    *fid* is the file to save the information to
    *fields* is the list of "step|time|value|point" fields to save

    The point field should be last in the list.
    """

    FIELDS = ["step", "time", "value", "point"]

    def __init__(self, problem, fid, fields=FIELDS):
        if any(f not in self.FIELDS for f in fields):
            raise ValueError("invalid monitor field")
        self.fid = fid
        self.fields = fields
        self.problem = problem
        self._pattern = "%%(%s)s\n" % (")s %(".join(fields))
        fid.write("# " + " ".join(fields) + "\n")

    def config_history(self, history):
        history.requires(time=1, value=1, point=1, step=1)

    def __call__(self, history):
        point = " ".join("%.15g" % v for v in history.point[0])
        time = "%g" % history.time[0]
        step = "%d" % history.step[0]
        scale, _ = nllf_scale(self.problem)
        value = "%.15g" % (scale * history.value[0])
        out = self._pattern % dict(point=point, time=time, value=value, step=step)
        self.fid.write(out)


class MonitorRunner(object):
    """
    Adaptor which allows solvers to accept progress monitors.
    """

    def __init__(self, monitors, problem):
        if monitors is None:
            monitors = [ConsoleMonitor(problem)]
        self.monitors = monitors
        self.history = History(time=1, step=1, point=1, value=1, population_points=1, population_values=1)
        for M in self.monitors:
            M.config_history(self.history)
        self._start = perf_counter()

    def __call__(self, step, point, value, population_points=None, population_values=None):
        self.history.update(
            time=perf_counter() - self._start,
            step=step,
            point=point,
            value=value,
            population_points=population_points,
            population_values=population_values,
        )
        for M in self.monitors:
            M(self.history)


class FitBase(object):
    """
    FitBase defines the interface from bumps models to the various fitting
    engines available within bumps.

    Each engine is defined in its own class with a specific set of attributes
    and methods.

    The *name* attribute is the name of the optimizer.  This is just a simple
    string.

    The *settings* attribute is a list of pairs (name, default), where the
    names are defined as fields in FitOptions.  A best attempt should be
    made to map the fit options for the optimizer to the standard fit options,
    since each of these becomes a new command line option when running
    bumps.  If that is not possible, then a new option should be added
    to FitOptions.  A plugin architecture might be appropriate here, if
    there are reasons why specific problem domains might need custom fitters,
    but this is not yet supported.

    Each engine takes a fit problem in its constructor.

    The :meth:`solve` method runs the fit.  It accepts a
    monitor to track updates, a mapper to distribute work and
    key-value pairs defining the settings.

    There are a number of optional methods for the fitting engines.  Basically,
    all the methods in :class:`FitDriver` first check if they are specialized
    in the fit engine before performing a default action.

    The *load*/*save* methods load and save the fitter state in a given
    directory with a specific base file name.  The fitter can choose a file
    extension to add to the base name.  Some care is needed to be sure that
    the extension doesn't collide with other extensions such as .mon for
    the fit monitor.

    The *plot* method shows any plots to help understand the performance of
    the fitter, such as a convergence plot showing the the range of values
    in the population over time, as well as plots of the parameter uncertainty
    if available.  The plot should work within  is given a figure canvas to work with

    The *stderr*/*cov* methods should provide summary statistics for the
    parameter uncertainties.  Some fitters, such as MCMC, will compute these
    directly from the population.  Others, such as BFGS, will produce an
    estimate of the uncertainty as they go along.  If the fitter does not
    provide these estimates, then they will be computed from numerical
    derivatives at the minimum in the FitDriver method.
    """

    def __init__(self, problem):
        """Fit the models and show the results"""
        self.problem = problem

    def solve(self, monitors=None, mapper=None, **options):
        raise NotImplementedError()


class MultiStart(FitBase):
    """
    Multi-start monte carlo fitter.

    This fitter wraps a local optimizer, restarting it a number of times
    to give it a chance to find a different local minimum.  If the keep_best
    option is True, then restart near the best fit, otherwise restart at
    random.
    """

    name = "Multistart Monte Carlo"
    settings = [("starts", 100)]

    def __init__(self, fitter):
        FitBase.__init__(self, fitter.problem)
        self.fitter = fitter

    def solve(self, monitors=None, mapper=None, **options):
        # TODO: need better way of tracking progress
        import logging

        starts = options.pop("starts", 1)
        reset = not options.pop("keep_best", True)
        f_best = np.inf
        x_best = self.problem.getp()
        for _ in range(max(starts, 1)):
            logging.info("multistart round %d", _)
            x, fx = self.fitter.solve(monitors=monitors, mapper=mapper, **options)
            if fx < f_best:
                x_best, f_best = x, fx
                logging.info("multistart f(x),x: %s %s", str(fx), str(x_best))
            if reset:
                self.problem.randomize()
            else:
                # Jitter
                self.problem.setp(x_best)
                pop = initpop.eps_init(1, self.problem.getp(), self.problem.bounds(), use_point=False, eps=1e-3)
                self.problem.setp(pop[0])
        return x_best, f_best


class DEFit(FitBase):
    """
    Classic Storn and Price differential evolution optimizer.
    """

    name = "Differential Evolution"
    id = "de"
    settings = [
        ("steps", 1000),
        ("pop", 10),
        ("CR", 0.9),
        ("F", 2.0),
        ("ftol", 1e-8),
        ("xtol", 1e-6),  # ('stop', ''),
    ]

    def solve(self, monitors=None, abort_test=None, mapper=None, **options):
        if abort_test is None:
            abort_test = lambda: False
        options = _fill_defaults(options, self.settings)
        from .mystic.optimizer import de
        from .mystic.solver import Minimizer
        from .mystic import stop

        if monitors is None:
            monitors = [ConsoleMonitor(self.problem)]
        if mapper is not None:
            _mapper = lambda p, v: mapper(v)
        else:
            _mapper = lambda p, v: list(map(self.problem.nllf, v))
        resume = hasattr(self, "state")
        steps = options["steps"] + (self.state["step"][-1] if resume else 0)
        strategy = de.DifferentialEvolution(
            npop=options["pop"], CR=options["CR"], F=options["F"], crossover=de.c_bin, mutate=de.rand1u
        )
        success = parse_tolerance(options)
        failure = stop.Steps(steps)
        self.history = History()
        # Step adds to current step number if resume
        minimize = Minimizer(
            strategy=strategy,
            problem=self.problem,
            history=self.history,
            monitors=monitors,
            success=success,
            failure=failure,
        )
        if resume:
            self.history.restore(self.state)
        x = minimize(mapper=_mapper, abort_test=abort_test, resume=resume)
        # print(minimize.termination_condition())
        # with open("/tmp/evals","a") as fid:
        #   print >>fid,minimize.history.value[0],minimize.history.step[0],\
        #       minimize.history.step[0]*options['pop']*len(self.problem.getp())
        return x, self.history.value[0]

    def load(self, input_path):
        self.state = load_history(input_path)

    def save(self, output_path):
        save_history(output_path, self.history.snapshot())


def parse_tolerance(options):
    from .mystic import stop

    if options.get("stop", ""):
        return stop.parse_condition(options["stop"])

    xtol, ftol = options["xtol"], options["ftol"]
    if xtol == 0:
        if ftol == 0:
            return None
        if ftol < 0:
            return stop.Rf(-ftol, scaled=True)
        return stop.Rf(ftol, scaled=False)
    else:
        if xtol == 0:
            return None
        if xtol < 0:
            return stop.Rx(-xtol, scaled=True)
        return stop.Rx(xtol, scaled=False)


def _history_file(path):
    return path + "-history.json"


def load_history(path):
    """
    Load fitter details from a history file.
    """
    import json

    with open(_history_file(path), "r") as fid:
        return json.load(fid)


def save_history(path, state):
    """
    Save fitter details to a history file as JSON.

    The content of the details are fitter specific.
    """
    import json

    with open(_history_file(path), "w") as fid:
        json.dump(state, fid)


class BFGSFit(FitBase):
    """
    BFGS quasi-newton optimizer.

    BFGS estimates Hessian and its Cholesky decomposition, but initial
    tests give uncertainties quite different from the directly computed
    Jacobian in Levenburg-Marquardt or the Hessian estimated at the
    minimum by numerical differentiation.

    To use the internal 'H' and 'L' and save some computation time, then
    use::

        C = lsqerror.chol_cov(fit.result['L'])
        stderr = lsqerror.stderr(C)
    """

    name = "Quasi-Newton BFGS"
    id = "newton"
    settings = [("steps", 3000), ("starts", 1), ("ftol", 1e-6), ("xtol", 1e-12)]

    def solve(self, monitors=None, abort_test=None, mapper=None, **options):
        if abort_test is None:
            abort_test = lambda: False
        options = _fill_defaults(options, self.settings)
        from .quasinewton import quasinewton

        self._update = MonitorRunner(problem=self.problem, monitors=monitors)
        result = quasinewton(
            fn=self.problem.nllf,
            x0=self.problem.getp(),
            monitor=self._monitor,
            abort_test=abort_test,
            itnlimit=options["steps"],
            gradtol=options["ftol"],
            steptol=1e-12,
            macheps=1e-8,
            eta=1e-8,
        )
        self.result = result
        # code = result['status']
        # from .quasinewton import STATUS
        # print("%d: %s, x=%s, fx=%s"
        #      % (code, STATUS[code], result['x'], result['fx']))
        return result["x"], result["fx"]

    def _monitor(self, step, x, fx):
        self._update(step=step, point=x, value=fx, population_points=[x], population_values=[fx])
        return True


class PSFit(FitBase):
    """
    Particle swarm optimizer.
    """

    name = "Particle Swarm"
    id = "ps"
    settings = [("steps", 3000), ("pop", 1)]

    def solve(self, monitors=None, mapper=None, **options):
        options = _fill_defaults(options, self.settings)
        if mapper is None:
            mapper = lambda x: list(map(self.problem.nllf, x))
        from .random_lines import particle_swarm

        self._update = MonitorRunner(problem=self.problem, monitors=monitors)
        low, high = self.problem.bounds()
        cfo = dict(
            parallel_cost=mapper, n=len(low), x0=self.problem.getp(), x1=low, x2=high, f_opt=0, monitor=self._monitor
        )
        npop = int(cfo["n"] * options["pop"])

        result = particle_swarm(cfo, npop, maxiter=options["steps"])
        satisfied_sc, n_feval, f_best, x_best = result

        return x_best, f_best

    def _monitor(self, step, x, fx, k):
        self._update(step=step, point=x[:, k], value=fx[k], population_points=x.T, population_values=fx)
        return True


class RLFit(FitBase):
    """
    Random lines optimizer.
    """

    name = "Random Lines"
    id = "rl"
    settings = [("steps", 3000), ("starts", 20), ("pop", 0.5), ("CR", 0.9)]

    def solve(self, monitors=None, abort_test=None, mapper=None, **options):
        if abort_test is None:
            abort_test = lambda: False
        options = _fill_defaults(options, self.settings)
        if mapper is None:
            mapper = lambda x: list(map(self.problem.nllf, x))
        from .random_lines import random_lines

        self._update = MonitorRunner(problem=self.problem, monitors=monitors)
        low, high = self.problem.bounds()
        cfo = dict(
            parallel_cost=mapper, n=len(low), x0=self.problem.getp(), x1=low, x2=high, f_opt=0, monitor=self._monitor
        )
        npop = max(int(cfo["n"] * options["pop"]), 3)

        result = random_lines(cfo, npop, abort_test=abort_test, maxiter=options["steps"], CR=options["CR"])
        satisfied_sc, n_feval, f_best, x_best = result

        return x_best, f_best

    def _monitor(self, step, x, fx, k):
        # print "rl best",k, x.shape,fx.shape
        self._update(step=step, point=x[:, k], value=fx[k], population_points=x.T, population_values=fx)
        return True


class PTFit(FitBase):
    """
    Parallel tempering optimizer.
    """

    name = "Parallel Tempering"
    id = "pt"
    settings = [("steps", 400), ("nT", 24), ("CR", 0.9), ("burn", 100), ("Tmin", 0.1), ("Tmax", 10)]

    def solve(self, monitors=None, mapper=None, **options):
        options = _fill_defaults(options, self.settings)
        # TODO: no mapper??
        from .partemp import parallel_tempering

        self._update = MonitorRunner(problem=self.problem, monitors=monitors)
        t = np.logspace(np.log10(options["Tmin"]), np.log10(options["Tmax"]), options["nT"])
        history = parallel_tempering(
            nllf=self.problem.nllf,
            p=self.problem.getp(),
            bounds=self.problem.bounds(),
            # logfile="partemp.dat",
            T=t,
            CR=options["CR"],
            steps=options["steps"],
            burn=options["burn"],
            monitor=self._monitor,
        )
        return history.best_point, history.best

    def _monitor(self, step, x, fx, P, E):
        self._update(step=step, point=x, value=fx, population_points=P, population_values=E)
        return True


class SimplexFit(FitBase):
    """
    Nelder-Mead simplex optimizer.
    """

    name = "Nelder-Mead Simplex"
    id = "amoeba"
    settings = [("steps", 1000), ("starts", 1), ("radius", 0.15), ("xtol", 1e-6), ("ftol", 1e-8)]

    def solve(self, monitors=None, abort_test=None, mapper=None, **options):
        from .simplex import simplex

        if abort_test is None:
            abort_test = lambda: False
        options = _fill_defaults(options, self.settings)
        # TODO: no mapper??
        self._update = MonitorRunner(problem=self.problem, monitors=monitors)
        # print("bounds", self.problem.bounds())
        result = simplex(
            f=self.problem.nllf,
            x0=self.problem.getp(),
            bounds=self.problem.bounds(),
            abort_test=abort_test,
            update_handler=self._monitor,
            maxiter=options["steps"],
            radius=options["radius"],
            xtol=options["xtol"],
            ftol=options["ftol"],
        )
        # Let simplex propose the starting point for the next amoeba
        # fit in a multistart amoeba context.  If the best is always
        # used, the fit can get stuck in a local minimum.
        self.problem.setp(result.next_start)
        # print("amoeba %s %s"%(result.x,result.fx))
        return result.x, result.fx

    def _monitor(self, k, n, x, fx):
        self._update(step=k, point=x[0], value=fx[0], population_points=x, population_values=fx)
        return True


class MPFit(FitBase):
    """
    MPFit optimizer.
    """

    name = "Levenberg-Marquardt"
    id = "lm"
    settings = [("steps", 200), ("ftol", 1e-10), ("xtol", 1e-10)]

    def solve(self, monitors=None, abort_test=None, mapper=None, **options):
        from .mpfit import mpfit

        if abort_test is None:
            abort_test = lambda: False
        options = _fill_defaults(options, self.settings)
        self._low, self._high = self.problem.bounds()
        self._update = MonitorRunner(problem=self.problem, monitors=monitors)
        self._abort = abort_test
        x0 = self.problem.getp()
        parinfo = []
        for low, high in zip(*self.problem.bounds()):
            parinfo.append(
                {
                    #'value': None,  # passed in by xall instead
                    #'fixed': False,  # everything is varying
                    "limited": (np.isfinite(low), np.isfinite(high)),
                    "limits": (low, high),
                    #'parname': '',  # could probably ask problem for this...
                    # From the code, default step size is sqrt(eps)*abs(value)
                    # or eps if value is 0.  This seems okay.  The other
                    # other alternative is to limit it by bounds.
                    #'step': 0,  # compute step automatically
                    #'mpside': 0,  # 1, -1 or 2 for right-, left- or 2-sided deriv
                    #'mpmaxstep': 0.,  # max step for this parameter
                    #'tied': '',  # parameter expressions tying fit parameters
                    #'mpprint': 1,  # print the parameter value when iterating
                }
            )

        result = mpfit(
            fcn=self._residuals,
            xall=x0,
            parinfo=parinfo,
            autoderivative=True,
            fastnorm=True,
            double=0,  # use single precision machine epsilon for derivative step
            # damp=0,  # no damping when damp=0
            # Stopping conditions
            ftol=options["ftol"],
            xtol=options["xtol"],
            # gtol=1e-100, # exclude gtol test
            maxiter=options["steps"],
            # Progress monitor
            iterfunct=self._monitor,
            nprint=1,  # call monitor each iteration
            quiet=True,  # leave it to monitor to print any info
            # Returns values
            nocovar=True,  # use our own covar calculation for consistency
        )

        if result.status > 0:
            x, fx = result.params, result.fnorm
        else:
            x, fx = None, None

        return x, fx

    def _monitor(self, fcn, p, k, fnorm, functkw=None, parinfo=None, quiet=0, dof=None, **extra):
        self._update(k, p, fnorm)

    def _residuals(self, p, fjac=None):
        if self._abort():
            return -1, None

        self.problem.setp(p)
        # treat prior probabilities on the parameters as additional
        # measurements
        residuals = np.hstack((self.problem.residuals().flat, self.problem.parameter_residuals()))
        # Tally costs for broken constraints
        extra_cost, failing_constraints = self.problem.constraints_nllf()
        # Spread the cost over the residuals.  Since we are smoothly increasing
        # residuals as we leave the boundary, this should push us back into the
        # boundary (within tolerance) during the lm fit.
        residuals += np.sign(residuals) * (extra_cost / len(residuals))
        return 0, residuals


class LevenbergMarquardtFit(FitBase):
    """
    Levenberg-Marquardt optimizer.
    """

    name = "Levenberg-Marquardt (scipy.leastsq)"
    id = "scipy.leastsq"
    settings = [("steps", 200), ("ftol", 1.5e-8), ("xtol", 1.5e-8)]
    # LM also has
    #    gtol: orthoganality between jacobian columns
    #    epsfcn: numerical derivative step size
    #    factor: initial radius
    #    diag: variable scale factors to bring them near 1

    def solve(self, monitors=None, abort_test=None, mapper=None, **options):
        from scipy import optimize

        if abort_test is None:
            abort_test = lambda: False
        options = _fill_defaults(options, self.settings)
        self._low, self._high = self.problem.bounds()
        self._update = MonitorRunner(problem=self.problem, monitors=monitors)
        x0 = self.problem.getp()
        maxfev = options["steps"] * (len(x0) + 1)
        result = optimize.leastsq(
            self._bounded_residuals,
            x0,
            ftol=options["ftol"],
            xtol=options["xtol"],
            maxfev=maxfev,
            epsfcn=1e-8,
            full_output=True,
        )
        x, cov_x, info, mesg, success = result
        if not 1 <= success <= 4:
            # don't treat "reached maxfev" as a true failure
            if "reached maxfev" in mesg:
                # unless the x values are bad
                if not np.all(np.isfinite(x)):
                    x = None
                    mesg = "Levenberg-Marquardt fit failed with bad values"
            else:
                x = None
        self._cov = cov_x if x is not None else None
        # compute one last time with x forced inside the boundary, and using
        # problem.nllf as returned by other optimizers.  We will ignore the
        # covariance output and calculate it again ourselves.  Not ideal if
        # f is expensive, but it will be consistent with other optimizers.
        if x is not None:
            x += self._stray_delta(x)
            self.problem.setp(x)
            fx = self.problem.nllf()
        else:
            fx = None
        return x, fx

    def _bounded_residuals(self, p):
        # Force the fit point into the valid region
        stray = self._stray_delta(p)
        stray_cost = np.sum(stray**2)
        if stray_cost > 0:
            stray_cost += 1e6
        self.problem.setp(p + stray)
        # treat prior probabilities on the parameters as additional
        # measurements
        residuals = np.hstack((self.problem.residuals().flat, self.problem.parameter_residuals()))
        # Tally costs for straying outside the boundaries plus other costs
        constraints_cost, failing_constraints = self.problem.constraints_nllf()
        extra_cost = stray_cost + constraints_cost
        # Spread the cost over the residuals.  Since we are smoothly increasing
        # residuals as we leave the boundary, this should push us back into the
        # boundary (within tolerance) during the lm fit.
        residuals += np.sign(residuals) * (extra_cost / len(residuals))
        return residuals

    def _stray_delta(self, p):
        """calculate how far point is outside the boundary"""
        return np.where(p < self._low, self._low - p, 0) + np.where(p > self._high, self._high - p, 0)

    def cov(self):
        return self._cov


class SnobFit(FitBase):
    name = "SNOBFIT"
    id = "snobfit"
    settings = [("steps", 200)]

    def solve(self, monitors=None, mapper=None, **options):
        options = _fill_defaults(options, self.settings)
        # TODO: no mapper??
        from snobfit.snobfit import snobfit

        self._update = MonitorRunner(problem=self.problem, monitors=monitors)
        x, fx, _ = snobfit(self.problem, self.problem.getp(), self.problem.bounds(), fglob=0, callback=self._monitor)
        return x, fx

    def _monitor(self, k, x, fx, improved):
        # TODO: snobfit does have a population...
        self._update(step=k, point=x, value=fx, population_points=[x], population_values=[fx])


class DreamModel(MCMCModel):
    """
    DREAM wrapper for fit problems.
    """

    def __init__(self, problem=None, mapper=None):
        """
        Create a sampling from the multidimensional likelihood function
        represented by the problem set using dream.
        """
        # print "dream"
        self.problem = problem
        self.bounds = self.problem.bounds()
        self.labels = self.problem.labels()

        self.mapper = mapper if mapper else lambda p: list(map(self.nllf, p))

    def log_density(self, x):
        return -self.nllf(x)

    def nllf(self, x):
        """Negative log likelihood of seeing models given *x*"""
        # Note: usually we will be going through the provided mapper, and
        # this function will never be called.
        # print "eval",x; sys.stdout.flush()
        return self.problem.nllf(x)

    def map(self, pop):
        # print "calling mapper",self.mapper
        return -np.array(self.mapper(pop))


class DreamFit(FitBase):
    name = "DREAM"
    id = "dream"
    settings = [
        ("samples", int(1e4)),
        ("burn", 100),
        ("pop", 10),
        ("init", "eps"),
        ("thin", 1),
        ("alpha", 0.01),
        ("outliers", "none"),
        ("trim", False),
        ("steps", 0),  # deprecated: use --samples instead
    ]

    def __init__(self, problem):
        FitBase.__init__(self, problem)
        self.dream_model = DreamModel(problem)
        self.state = None

    def solve(self, monitors=None, abort_test=None, mapper=None, **options):
        from .dream import Dream

        if abort_test is None:
            abort_test = lambda: False
        options = _fill_defaults(options, self.settings)

        if mapper:
            self.dream_model.mapper = mapper
        self._update = MonitorRunner(problem=self.dream_model.problem, monitors=monitors)

        population = initpop.generate(self.dream_model.problem, **options)
        pop_size = population.shape[0]
        draws, steps = int(options["samples"]), options["steps"]
        if steps == 0:
            steps = (draws + pop_size - 1) // pop_size
        # TODO: need a better way to announce number of steps
        # maybe somehow print iteration # of # iters in the monitor?
        print("# steps: %d, # draws: %d" % (steps, pop_size * steps))
        population = population[None, :, :]
        sampler = Dream(
            model=self.dream_model,
            population=population,
            draws=pop_size * steps,
            burn=pop_size * options["burn"],
            thinning=options["thin"],
            monitor=self._monitor,
            alpha=options["alpha"],
            outlier_test=options["outliers"],
            DE_noise=1e-6,
        )

        self.state = sampler.sample(state=self.state, abort_test=abort_test)

        self._trimmed = self.state.trim_portion() if options["trim"] else 1.0
        # print("trimming", options['trim'], self._trimmed)
        self.state.mark_outliers(portion=self._trimmed)
        self.state.keep_best()
        self.state.title = self.dream_model.problem.name

        # TODO: Temporary hack to apply a post-mcmc action to the state vector
        # The problem is that if we manipulate the state vector before saving
        # it then we will not be able to use the --resume feature.  We can
        # get around this by just not writing state for the derived variables,
        # at which point we can remove this notice.
        # TODO: Add derived/visible variable support to other optimizers
        fn, labels = getattr(self.problem, "derive_vars", (None, None))
        if fn is not None:
            self.state.derive_vars(fn, labels=labels)
        visible_vars = getattr(self.problem, "visible_vars", None)
        if visible_vars is not None:
            self.state.set_visible_vars(visible_vars)
        integer_vars = getattr(self.problem, "integer_vars", None)
        if integer_vars is not None:
            self.state.set_integer_vars(integer_vars)

        x, fx = self.state.best()

        # Check that the last point is the best point
        # points, logp = self.state.sample()
        # assert logp[-1] == fx
        # print(points[-1], x)
        # assert all(points[-1, i] == xi for i, xi in enumerate(x))
        return x, -fx

    def entropy(self, **kw):
        return self.state.entropy(portion=self._trimmed, **kw)

    def _monitor(self, state, pop, logp):
        # Get an early copy of the state
        self.state = self._update.history.uncertainty_state = state
        step = state.generation
        x, fx = state.best()
        self._update(step=step, point=x, value=-fx, population_points=pop, population_values=-logp)
        return True

    def stderr(self):
        """
        Approximate standard error as 1/2 the 68% interval fo the sample,
        which is a more robust measure than the mean of the sample for
        non-normal distributions.
        """
        from .dream.stats import var_stats

        vstats = var_stats(self.state.draw(portion=self._trimmed))
        return np.array([(v.p68[1] - v.p68[0]) / 2 for v in vstats], "d")

    # def cov(self):
    #    # Covariance estimate from final 1000 points
    #    return np.cov(self.state.draw().points[-1000:])

    def load(self, input_path):
        from .dream.state import load_state, path_contains_saved_state

        if path_contains_saved_state(input_path):
            print("loading saved state from %s (this might take awhile) ..." % (input_path,))
            fn, labels = getattr(self.problem, "derive_vars", (None, []))
            self.state = load_state(input_path, report=100, derived_vars=len(labels))
        else:
            # Warn if mc files are not found on --resume path
            warnings.warn("No mcmc found; ignoring --resume=%r" % input_path)

    def save(self, output_path):
        self.state.save(output_path)

    def plot(self, output_path):
        self.state.show(figfile=output_path, portion=self._trimmed)
        self.error_plot(figfile=output_path)

    def show(self):
        pass

    def error_plot(self, figfile):
        # Produce error plot
        import pylab
        from . import errplot

        # TODO: shouldn't mix calc and display!
        res = errplot.calc_errors_from_state(problem=self.dream_model.problem, state=self.state, portion=self._trimmed)
        if res is not None:
            pylab.figure()
            errplot.show_errors(res)
            pylab.savefig(figfile + "-errors.png", format="png")


class Resampler(FitBase):
    # TODO: why isn't cli.resynth using this?

    def __init__(self, fitter):
        self.fitter = fitter
        raise NotImplementedError()

    def solve(self, **options):
        starts = options.pop("starts", 1)
        restart = options.pop("restart", False)
        x, fx = self.fitter.solve(**options)
        points = _resampler(self.fitter, x, samples=starts, restart=restart, **options)
        self.points = points  # save points for later plotting
        return x, fx


def _resampler(fitter, xinit, samples=100, restart=False, **options):
    """
    Refit the result multiple times with resynthesized data, building
    up an array in Result.samples which contains the best fit to the
    resynthesized data.  *samples* is the number of samples to generate.
    *fitter* is the (local) optimizer to use. **kw are the parameters
    for the optimizer.
    """
    x = xinit
    points = []
    try:  # TODO: some solvers already catch KeyboardInterrupt
        for _ in range(samples):
            # print "== resynth %d of %d" % (i, samples)
            fitter.problem.resynth_data()
            if restart:
                fitter.problem.randomize()
            else:
                fitter.problem.setp(x)
            x, fx = fitter.solve(**options)
            points.append(np.hstack((fx, x)))
            # print self.problem.summarize()
            # print "[chisq=%g]" % (nllf*2/self.problem.dof)
    except KeyboardInterrupt:
        # On keyboard interrupt we can declare that we are finished sampling
        # without it being an error condition, so let this exception pass.
        pass
    finally:
        # Restore the state of the problem
        fitter.problem.restore_data()
        fitter.problem.setp(xinit)
        # fitter.problem.model_update()  # setp does model update
    return points


class FitDriver(object):
    def __init__(self, fitclass=None, problem=None, monitors=None, abort_test=None, mapper=None, **options):
        self.fitclass = fitclass
        self.problem = problem
        self.options = options
        self.monitors = monitors
        self.abort_test = abort_test
        self.mapper = mapper if mapper else lambda p: list(map(problem.nllf, p))
        self.fitter = None
        self.result = None

    def fit(self, resume=None):
        if hasattr(self, "_cov"):
            del self._cov
        if hasattr(self, "_stderr"):
            del self._stderr
        fitter = self.fitclass(self.problem)
        if resume:
            fitter.load(resume)
        starts = self.options.get("starts", 1)
        if starts > 1:
            fitter = MultiStart(fitter)
        t0 = perf_counter()
        self.fitter = fitter
        x, fx = fitter.solve(monitors=self.monitors, abort_test=self.abort_test, mapper=self.mapper, **self.options)
        self.time = perf_counter() - t0
        self.result = x, fx
        if x is not None:
            self.problem.setp(x)
        return x, fx

    def clip(self):
        """
        Force parameters within bounds so constraints are finite.

        The problem is updated with the new parameter values.

        Returns a list of parameter names that were clipped.
        """
        labels = self.problem.labels()
        values = self.problem.getp()
        bounds = self.problem.bounds()
        new_values = np.clip(values, bounds[0], bounds[1])
        clipped = [name for name, old, new in zip(labels, values, new_values) if old != new]
        self.problem.setp(new_values)
        return clipped

    def entropy(self, method=None):
        if hasattr(self.fitter, "entropy"):
            return self.fitter.entropy(method=method)
        else:
            from .dream import entropy

            return entropy.cov_entropy(self.cov()), 0

    def chisq(self):
        if not hasattr(self, "_chisq"):
            self._chisq = self.problem.chisq()
        return self._chisq

    def cov(self):
        r"""
        Return an estimate of the covariance of the fit.

        Depending on the fitter and the problem, this may be computed from
        existing evaluations within the fitter, or from numerical
        differentiation around the minimum.

        If the problem uses $\chi^2/2$ as its nllf, then the covariance
        is derived from the Jacobian::

            x = fit.problem.getp()
            J = bumps.lsqerror.jacobian(fit.problem, x)
            cov = bumps.lsqerror.jacobian_cov(J)

        Otherwise, the numerical differentiation will use the Hessian
        estimated from nllf::

            x = fit.problem.getp()
            H = bumps.lsqerror.hessian(fit.problem, x)
            cov = bumps.lsqerror.hessian_cov(H)
        """
        # Note: if fit() has not been run then self.fitter is None and in
        # particular, self.fitter will not have a covariance matrix.  In
        # this case, the code will fall through to computing the covariance
        # matrix directly from the problem.  It will use the initial value
        # stored in the problem parameters because results will also be None.
        if not hasattr(self, "_cov"):
            self._cov = None
            if hasattr(self.fitter, "cov"):
                self._cov = self.fitter.cov()
                # print("fitter cov", self._cov)
        if self._cov is None:
            # Use Jacobian if residuals are available because it is faster
            # to compute.  Otherwise punt and use Hessian.  The has_residuals
            # attribute should be True if present.  It may be false if
            # the problem defines a residuals method but doesn't really
            # have residuals (e.g. to allow levenberg-marquardt to run even
            # though it is not fitting a sum-square problem).
            if hasattr(self.problem, "has_residuals"):
                has_residuals = self.problem.has_residuals
            else:
                has_residuals = hasattr(self.problem, "residuals")
            x = self.problem.getp() if self.result is None else self.result[0]
            if has_residuals:
                J = lsqerror.jacobian(self.problem, x)
                # print("Jacobian", J)
                self._cov = lsqerror.jacobian_cov(J)
            else:
                H = lsqerror.hessian(self.problem, x)
                # print("Hessian", H)
                self._cov = lsqerror.hessian_cov(H)
        return self._cov

    def stderr(self):
        """
        Return an estimate of the standard error of the fit.

        Depending on the fitter and the problem, this may be computed from
        existing evaluations within the fitter, or from numerical
        differentiation around the minimum.
        """
        # Note: if fit() has not been run then self.fitter is None and in
        # particular, self.fitter will not have a stderr method defined so
        # it will compute stderr from covariance.
        if not hasattr(self, "_stderr"):
            self._stderr = None
            if hasattr(self.fitter, "stderr"):
                self._stderr = self.fitter.stderr()
        if self._stderr is None:
            # If no stderr from the fitter then compute it from the covariance
            self._stderr = self.stderr_from_cov()
        return self._stderr

    def stderr_from_cov(self):
        """
        Return an estimate of standard error of the fit from covariance matrix.

        Unlike stderr, which uses the estimate from the underlying
        fitter (DREAM uses the MCMC sample for this), *stderr_from_cov*
        estimates the error from the diagonal of the covariance matrix.
        Here, the covariance matrix may have been estimated by the fitter
        instead of the Hessian.
        """
        if not hasattr(self, "_stderr_from_cov"):
            self._stderr_from_cov = lsqerror.stderr(self.cov())
        return self._stderr_from_cov

    def show(self):
        if hasattr(self.fitter, "show"):
            self.fitter.show()
        if hasattr(self.problem, "show"):
            self.problem.show()

    # TODO: reenable the "implied variance" calculation
    def _unused_show_err(self):
        """
        Display the error approximation from the numerical derivative.

        Warning: cost grows as the cube of the number of parameters.
        """
        # TODO: need cheaper uncertainty estimate
        # Note: error estimated from hessian diagonal is insufficient.
        err = self.stderr_from_cov()
        # TODO: citation needed
        # The "implied variance" column is obtained by scaling the covariance
        # matrix so that chisq = DOF. Any excess chisq implies increased
        # variance in the measurements, so increased variance in the parameters.
        # This is well defined for linear systems with equal but unknown
        # variance in each measurement, and assumed to be approximately true
        # for nonlinear systems, with the unexplained variance distributed
        # proportionately amongst the measurement uncertainties.
        norm = np.sqrt(self.chisq())
        print("=== Uncertainty from curvature:     name value(unc.)     value(unc./chi)) ===")
        for k, v, dv in zip(self.problem.labels(), self.problem.getp(), err):
            print("%40s %-15s %-15s" % (k, format_uncertainty(v, dv), format_uncertainty(v, dv / norm)))
        print("=" * 75)

    def show_err(self):
        """
        Display the error approximation from the numerical derivative.

        Warning: cost grows as the cube of the number of parameters.
        """
        # TODO: need cheaper uncertainty estimate
        # Note: error estimated from hessian diagonal is insufficient.
        err = self.stderr_from_cov()
        print("=== Uncertainty from curvature:     name   value(unc.) ===")
        for k, v, dv in zip(self.problem.labels(), self.problem.getp(), err):
            print(f"{k:>40s}   {format_uncertainty(v, dv):<15s}")
        print("=" * 58)

    def show_cov(self):
        cov = self.cov()
        maxn = 1000  # max array dims to print
        cov_str = np.array2string(
            cov,
            max_line_width=20 * maxn,
            threshold=maxn * maxn,
            precision=6,  # suppress_small=True,
            separator=", ",
        )
        print("=== Covariance matrix ===")
        print(cov_str)
        print("=========================")

    def show_entropy(self, method=None):
        print("Calculating entropy...")
        S, dS = self.entropy(method=method)
        print("Entropy: %s bits" % format_uncertainty(S, dS))

    def save(self, output_path):
        # print "calling driver save"
        if hasattr(self.fitter, "save"):
            self.fitter.save(output_path)
        if hasattr(self.problem, "save"):
            self.problem.save(output_path)

    def load(self, input_path):
        # print "calling driver save"
        if hasattr(self.fitter, "load"):
            self.fitter.load(input_path)
        if hasattr(self.problem, "load"):
            self.problem.load(input_path)

    def plot(self, output_path, view=None):
        # print "calling fitter.plot"
        if hasattr(self.problem, "plot"):
            self.problem.plot(figfile=output_path, view=view)
        if hasattr(self.fitter, "plot"):
            self.fitter.plot(output_path=output_path)

    def _save_fit_cov(self, output_path):
        model = getattr(self.problem, "name", self.problem.__class__.__name__)
        fitter = self.fitclass.id
        cov = self.cov()
        err = self.stderr_from_cov()
        chisq = self.chisq()

        state = {
            "model": model,
            "fitter": fitter,
        }


def _fill_defaults(options, settings):
    """
    Returns options dict with missing values filled from settings.
    """
    result = dict(settings)  # settings is a list of (key,value) pairs
    result.update(options)
    return result


FITTERS = []
FIT_AVAILABLE_IDS = []
FIT_ACTIVE_IDS = []


def register(fitter, active=True):
    """
    Register a new fitter with bumps, if it is not already there.

    *active* is False if you don't want it showing up in the GUI selector.
    """
    # Check if already registered.
    if fitter in FITTERS:
        return

    # Check that there is no other fitter of that name
    if fitter.id in FIT_AVAILABLE_IDS:
        raise ValueError("There is already a fitter registered as %r" % fitter.id)

    # Register the fitter.
    FITTERS.append(fitter)
    FIT_AVAILABLE_IDS.append(fitter.id)

    # Make it "active" by listing it in the help menu.
    if active:
        FIT_ACTIVE_IDS.append(fitter.id)


# Register the fitters
register(SimplexFit, active=True)
register(DEFit, active=True)
register(DreamFit, active=True)
register(BFGSFit, active=True)
register(LevenbergMarquardtFit, active=True)
register(MPFit, active=True)
# register(PSFit, active=False)
register(PTFit, active=False)
# register(RLFit, active=False)
# register(SnobFit, active=False)

FIT_DEFAULT_ID = SimplexFit.id

assert FIT_DEFAULT_ID in FIT_ACTIVE_IDS
assert all(f in FIT_AVAILABLE_IDS for f in FIT_ACTIVE_IDS)


def fit(problem, method=FIT_DEFAULT_ID, verbose=False, **options):
    """
    Simplified fit interface.

    Given a fit problem, the name of a fitter and the fitter options,
    it will run the fit and return the best value and standard error of
    the parameters.  If *verbose* is true, then the console monitor will
    be enabled, showing progress through the fit and showing the parameter
    standard error at the end of the fit, otherwise it is completely
    silent.

    Returns an *OptimizeResult* object containing "x" and "dx".  The
    dream fitter also includes the "state" object, allowing for more
    detailed uncertainty analysis.  Optimizer information such as the
    stopping condition and the number of function evaluations are not
    yet included.

    To run in parallel (with multiprocessing and dream)::

        from bumps.mapper import MPMapper
        mapper = MPMapper.start_mapper(problem, None, cpu=0) #cpu=0 for all CPUs
        result = fit(problem, method="dream", mapper=mapper)

    """
    from scipy.optimize import OptimizeResult

    # verbose = True
    if method not in FIT_AVAILABLE_IDS:
        raise ValueError("unknown method %r not one of %s" % (method, ", ".join(sorted(FIT_ACTIVE_IDS))))
    for fitclass in FITTERS:
        if fitclass.id == method:
            break
    monitors = None if verbose else []  # default is step monitor
    driver = FitDriver(fitclass=fitclass, problem=problem, monitors=monitors, **options)
    driver.clip()  # make sure fit starts within domain
    x0 = problem.getp()
    x, fx = driver.fit()
    problem.setp(x)
    dx = driver.stderr()
    if verbose:
        print("final chisq", problem.chisq_str())
        driver.show_err()
    result = OptimizeResult(
        x=x,
        dx=driver.stderr(),
        fun=fx,
        success=True,
        status=0,
        message="successful termination",
        # nit=0, # number of iterations
        # nfev=0, # number of function evaluations
        # njev, nhev # jacobian and hessian evaluations
        # maxcv=0, # max constraint violation
    )
    if hasattr(driver.fitter, "state"):
        result.state = driver.fitter.state
    return result


def test_fitters():
    """
    Run the fit tests to make sure they work.
    """
    from .curve import Curve
    from .fitproblem import FitProblem

    x = [1, 2, 3, 4, 5, 6]
    y = [2.1, 4.0, 6.3, 8.03, 9.6, 11.9]
    dy = [0.05, 0.05, 0.2, 0.05, 0.2, 0.2]

    def line(x, m, b=0):
        return m * x + b

    M = Curve(line, x, y, dy, m=2, b=2)
    M.m.range(0, 4)
    M.b.range(-5, 5)

    problem = FitProblem(M)

    # Set the tolerance for the tests (relative)
    fit_value_tol = 1e-3
    fit_error_tol = 1e-3
    expected_value = [1.106e-1, 1.970]
    expected_error = [5.799e-2, 2.055e-2]

    for fitter_name in FIT_ACTIVE_IDS:
        result = fit(problem, method=fitter_name, verbose=False)
        assert np.allclose(result.x, expected_value, rtol=fit_value_tol)
        if fitter_name != "dream":
            # dream error bars vary too much to test
            assert np.allclose(result.dx, expected_error, rtol=fit_error_tol)



"""
Statistics helper functions.
"""

__all__ = [
    "VarStats",
    "var_stats",
    "format_vars",
    "parse_var",
    "stats",
    "credible_interval",
    "shortest_credible_interval",
]

import re
import json

import numpy as np

from .formatnum import format_uncertainty


class VarStats(object):
    def __init__(self, **kw):
        self.__dict__ = kw


def var_stats(draw, vars=None):
    if vars is None:
        vars = range(draw.points.shape[1])
    return [_var_stats_one(draw, v) for v in vars]


ONE_SIGMA = 1 - 2 * 0.15865525393145705


def _var_stats_one(draw, var):
    weights, values = draw.weights, draw.points[:, var].flatten()

    integer = draw.integers is not None and draw.integers[var]
    if integer:
        values = np.floor(values)

    best_idx = np.argmax(draw.logp)
    best = values[best_idx]
    x_sort_index = draw.get_argsort_indices(var)

    # Choose the interval for the histogram
    # credible_interval = shortest_credible_interval
    p95, p68, p0 = credible_interval(x=values, weights=weights, ci=[0.95, ONE_SIGMA, 0.0], x_sort_index=x_sort_index)

    ## reporting uncertainty on credible intervals?
    ## might be nice to pair sd on credible intervals
    ## with the actual CIs, rather than use a separate param
    # from .digits import credible_inderval_sd
    # p95sd = credible_interval_sd(values, 0.95)
    # p68sd = credible_interval_sd(values, ONE_SIGMA)

    # open('/tmp/out','a').write(
    #     "in vstats: p68=%s, p95=%s, p0=%s, value range=%s\n"
    #     % (p68,p95,p0,(min(values),max(values))))
    # if p0[0] != p0[1]: raise RuntimeError("wrong median %s"%(str(p0),))

    mean, std = stats(x=values, weights=weights, x_sort_index=x_sort_index)

    vstats = VarStats(
        label=draw.labels[var],
        index=var + 1,
        p95=p95,
        p95_range=(p95[0], p95[1] + integer * 0.9999999999),
        p68=p68,
        p68_range=(p68[0], p68[1] + integer * 0.9999999999),
        # p95sd=p95sd, p68sd=p68sd,
        median=p0[0],
        mean=mean,
        std=std,
        best=best,
        integer=integer,
    )

    return vstats


def format_num(x, place):
    precision = 10**place
    digits_after_decimal = abs(place) if place < 0 else 0
    return "%.*f" % (digits_after_decimal, np.round(x / precision) * precision)


def format_vars(all_vstats):
    v = dict(
        parameter="Parameter",
        mean="mean",
        median="median",
        best="best",
        interval68="68% interval",
        interval95="95% interval",
    )
    s = ["   %(parameter)20s %(mean)10s %(median)7s %(best)7s " "[%(interval68)15s] [%(interval95)15s]" % v]
    for v in all_vstats:
        # Make sure numbers are formatted with the appropriate precision
        place = (
            int(np.log10(v.p95[1] - v.p95[0])) - 2
            if v.p95[1] > v.p95[0]
            else int(np.log10(abs(v.p95[0]))) - 3
            if v.p95[0] != 0
            else 0
        )
        summary = dict(
            mean=format_uncertainty(v.mean, v.std),
            median=format_num(v.median, place - 1),
            best=format_num(v.best, place - 1),
            lo68=format_num(v.p68[0], place),
            hi68=format_num(v.p68[1], place),
            loci=format_num(v.p95[0], place),
            hici=format_num(v.p95[1], place),
            parameter=v.label,
            index=v.index,
        )
        s.append(
            "%(index)2d %(parameter)20s %(mean)10s %(median)7s %(best)7s "
            "[%(lo68)7s %(hi68)7s] [%(loci)7s %(hici)7s]" % summary
        )

    return "\n".join(s)


def save_vars(all_vstats, filename):
    with open(filename, "w") as fid:
        json.dump(
            dict((v.label, v.__dict__) for v in all_vstats),
            fid,
            default=numpy_json,
            sort_keys=True,
            indent=2,
        )


def numpy_json(o):
    """
    JSON encoder for numpy data.

    To automatically convert numpy data to lists when writing a datastream
    use json.dumps(object, default=numpy_json).
    """
    try:
        return o.tolist()
    except AttributeError:
        raise TypeError


VAR_PATTERN = re.compile(
    r"""
   ^\ *
   (?P<parnum>[0-9]+)\ +
   (?P<parname>.+?)\ +
   (?P<mean>[0-9.-]+?)
   \((?P<err>[0-9]+)\)
   (e(?P<exp>[+-]?[0-9]+))?\ +
   (?P<median>[0-9.eE+-]+?)\ +
   (?P<best>[0-9.eE+-]+?)\ +
   \[\ *(?P<lo68>[0-9.eE+-]+?)\ +
   (?P<hi68>[0-9.eE+-]+?)\]\ +
   \[\ *(?P<lo95>[0-9.eE+-]+?)\ +
   (?P<hi95>[0-9.eE+-]+?)\]
   \ *$
   """,
    re.VERBOSE,
)


def parse_var(line):
    """
    Parse a line returned by format_vars back into the statistics for the
    variable on that line.
    """
    m = VAR_PATTERN.match(line)
    if m:
        exp = int(m.group("exp")) if m.group("exp") else 0
        return VarStats(
            index=int(m.group("parnum")),
            name=m.group("parname"),
            mean=float(m.group("mean")) * 10**exp,
            median=float(m.group("median")),
            best=float(m.group("best")),
            p68=(float(m.group("lo68")), float(m.group("hi68"))),
            p95=(float(m.group("lo95")), float(m.group("hi95"))),
        )
    else:
        return None


def stats(x, weights=None, x_sort_index=None):
    """
    Find mean and standard deviation of a set of weighted samples.

    Note that the median is not strictly correct (we choose an endpoint
    of the sample for the case where the median falls between two values
    in the sample), but this is good enough when the sample size is large.
    """
    if weights is None:
        if x_sort_index is None:
            x_sort_index = np.argsort(x)
        x = x[x_sort_index]
        mean, std = np.mean(x), np.std(x, ddof=1)
    else:
        mean = np.mean(x * weights) / np.sum(weights)
        # TODO: this is biased by selection of mean; need an unbiased formula
        var = np.sum((weights * (x - mean)) ** 2) / np.sum(weights)
        std = np.sqrt(var)

    return mean, std


def credible_interval(x, ci, weights=None, x_sort_index=None):
    r"""
    Find the credible interval covering the portion *ci* of the data.

    *x* are samples from the posterior distribution.

    *ci* is a set of intervals in [0,1].  For a $1-\sigma$ interval use
    *ci=erf(1/sqrt(2))*, or 0.68. About 1e5 samples are needed for 2 digits
    of  precision on a $1-\sigma$ credible interval.  For a 95% interval,
    about 1e6 samples are needed for 2 digits of precision.  At least 1000
    points are needed for an unbiased result, otherwise the resulting interval
    will be shorter than expected (tested on a variety of distributions
    including exponential, cauchy, gaussian, beta and gamma).

    *weights* is a vector of weights for each x, or None for unweighted.
    One could weight points according to temperature in a parallel tempering
    dataset.

    Returns an array *[[x1_low, x1_high], [l2_low, x2_high], ...]* where
    *[xi_low, xi_high]* are the starting and ending values for credible
    interval *i*.

    This function is faster if the inputs are already sorted.
    """
    n = x.size
    ci = np.asarray(ci, "d")
    target = (1 + np.vstack((-ci, +ci))).T / 2
    if x_sort_index is None:
        x_sort_index = np.argsort(x)

    if weights is None:
        cdf = np.linspace(0.5 / n, 1 - 0.5 / n, n)
        # cdf = np.linspace(1, n, n)/(n+1)
        result = np.interp(target, cdf, x[x_sort_index])
    else:
        x, weights = x[x_sort_index], weights[x_sort_index]
        # convert weights to cdf
        cdf = np.cumsum(weights)
        cdf /= cdf[-1]
        cdf -= 0.5 * cdf[0]
        # cdf *= n/(cdf[-1]*(n+1))
        result = np.interp(target, cdf, x)
    return result if ci.shape else result[0]


def shortest_credible_interval(x, ci=0.95, weights=None):
    """
    Find the credible interval covering the portion *ci* of the data.

    *x* are samples from the posterior distribution.
    *ci* is the interval size in (0,1], and defaults to 0.95.
    For a 1-sigma interval use *ci=erf(1/sqrt(2))*.
    *weights* is a vector of weights for each x, or None for unweighted.

    Returns the minimum and maximum values of the interval.
    If *ci* is a vector, return a vector of intervals.

    This function is faster if the inputs are already sorted.

    About 1e6 samples are needed for 2 digits of precision on a 95%
    credible interval, or 1e5 for 2 digits on a 1-sigma credible interval.

    To remove bias towards toward smaller intervals, the midpoints between
    the surrounding intervals are used as the end points.
    """

    if weights is None:
        x = np.sort(x)
        # Simple solution: ci*N is the number of points in the interval, so
        # find the width of every interval of that size and return the smallest.
        if np.isscalar(ci):
            return _unweighted_hpd(x, ci)
        else:
            return [_unweighted_hpd(x, ci_k) for ci_k in ci]
    else:
        index = np.argsort(x)
        x, weights = x[index], weights[index]
        # Work from the empirical cdf, finding the corresponding right
        # interval for each possible left interval and choosing that with
        # the shortest distance.
        cdf = np.cumsum(weights)
        cdf /= cdf[-1]
        # jcdf -= 0.5*cdf[0]
        if np.isscalar(ci):
            return _weighted_hpd(x, cdf, ci)
        else:
            return [_weighted_hpd(x, cdf, ci_k) for ci_k in ci]


def _unweighted_hpd(x, ci):
    """
    Find shortest credible interval ci in sorted, unweighted x
    """
    n = len(x)
    size = int(ci * n)
    if size >= n:
        return x[0], x[-1]
    else:
        width = x[size:] - x[:-size]
        index = np.argmin(width)
        # left, right = x[idx], x[idx+size]
        left = x[0] if index == 0 else (x[index - 1] + x[index]) / 2
        right = x[-1] if index + size == n - 1 else (x[index + size] + x[index + size + 1]) / 2
        return left, right


def _weighted_hpd(z, cdf, ci):  # extra one-half interval
    """
    Find shortest credible interval ci in sorted, weighted x
    """
    size = np.searchsorted(cdf, 1 - ci)
    if size == 0:
        return z[0], z[-1]
    p_left = cdf[:size]
    z_left = z[:size]
    # avoid spurious floating point bugs, e.g., where .1+0.9 > 1.0
    i_right = np.searchsorted(cdf[:-1], p_left + ci)
    z_right = z[i_right]
    index = np.argmin(z_right - z_left)
    left = z_left[0] if index == 0 else (z_left[index - 1] + z_left[index]) / 2
    right = z_right[-1] if index + 1 == len(z_right) else (z_right[index] + z_right[index + 1]) / 2
    return left, right



"""
Build layout for histogram plots
"""

__all__ = ["var_plot_size", "plot_vars", "plot_var"]

from math import ceil, sqrt

import numpy as np
from matplotlib import pyplot as plt

# Set space between plots in horiz and vert.
H_SPACE = 0.2
V_SPACE = 0.2

# Set top, bottom, left margins.
T_MARGIN = 0.2
B_MARGIN = 0.2
L_MARGIN = 0.2
R_MARGIN = 0.4

# Set desired plot sizes.
TILE_W = 3.0
TILE_H = 2.0
CBAR_WIDTH = 0.75


def var_plot_size(n):
    ncol, nrow = tile_axes_square(n)

    # Calculate total width and figure size
    plots_width = (TILE_W + H_SPACE) * ncol
    figwidth = plots_width + CBAR_WIDTH + L_MARGIN + R_MARGIN
    figheight = (TILE_H + V_SPACE) * nrow + T_MARGIN + B_MARGIN
    return figwidth, figheight


def _make_var_axes(n, fig=None):
    """
    Build a figure with one axis per parameter,
    and one axis (the last one) to contain the colorbar.
    Use to make the vars histogram figure.
    """
    if fig is None:
        fig = plt.gcf()
    fig.clf()
    total_width, total_height = fig.get_size_inches()

    ncol, nrow = tile_axes_square(n)

    # Calculate dimensions as a faction of figure size.
    v_space_f = V_SPACE / total_height
    h_space_f = H_SPACE / total_width
    t_margin_f = T_MARGIN / total_height
    b_margin_f = B_MARGIN / total_height
    l_margin_f = L_MARGIN / total_width
    top = 1 - t_margin_f + v_space_f
    left = l_margin_f

    tile_h = (total_height - T_MARGIN - B_MARGIN) / nrow - V_SPACE
    tile_w = (total_width - L_MARGIN - R_MARGIN - CBAR_WIDTH) / ncol - H_SPACE
    tile_h_f = tile_h / total_height
    tile_w_f = tile_w / total_width

    # Calculate colorbar location (left, bottom) and colorbar height.
    l_cbar_f = l_margin_f + ncol * (tile_w_f + h_space_f)
    b_cbar_f = b_margin_f + v_space_f
    cbar_w_f = CBAR_WIDTH / total_width
    cbar_h_f = 1 - t_margin_f - b_margin_f - v_space_f
    cbar_box = [l_cbar_f, b_cbar_f, cbar_w_f, cbar_h_f]

    k = 0
    for j in range(1, nrow + 1):
        for i in range(0, ncol):
            if k >= n:
                break
            dims = [left + i * (tile_w_f + h_space_f), top - j * (tile_h_f + v_space_f), tile_w_f, tile_h_f]
            ax = fig.add_axes(dims)
            ax.set_facecolor("none")
            k += 1

    fig.add_axes(cbar_box)
    # fig.set_size_inches(total_width, total_height)
    return fig


def tile_axes_square(n):
    """
    Determine number of columns by finding the
    next greatest square, then determine number
    of rows needed.
    """
    cols = int(ceil(sqrt(n)))
    rows = int(ceil(n / float(cols)))
    return cols, rows


def plot_vars(draw, all_vstats, fig=None, **kw):
    n = len(all_vstats)
    fig = _make_var_axes(n, fig=fig)
    cbar = _make_fig_colorbar(draw.logp, fig=fig)
    for k, vstats in enumerate(all_vstats):
        axes = fig.axes[k]
        plot_var(draw, vstats, k, cbar, axes=axes, **kw)
        fig.canvas.draw()


def plot_var(draw, vstats, var, cbar, nbins=30, axes=None):
    values = draw.points[:, var].flatten()
    bin_range = vstats.p95_range
    # bin_range = np.min(values), np.max(values)
    import pylab

    if axes is None:
        axes = pylab.gca()

    _make_logp_histogram(values, draw.logp, nbins, bin_range, draw.weights, cbar, axes)
    _decorate_histogram(vstats, axes)


def _decorate_histogram(vstats, axes):
    from matplotlib.transforms import blended_transform_factory as blend

    l95, h95 = vstats.p95_range
    l68, h68 = vstats.p68_range

    # Shade things inside 1-sigma
    axes.axvspan(l68, h68, color="gold", alpha=0.5, zorder=-1)
    # build transform with x=data, y=axes(0,1)
    transform = blend(axes.transData, axes.transAxes)

    def marker(symbol, position):
        if position < l95:
            symbol, position, ha = "<" + symbol, l95, "left"
        elif position > h95:
            symbol, position, ha = ">" + symbol, h95, "right"
        else:
            symbol, position, ha = symbol, position, "center"
        axes.text(position, 0.95, symbol, va="top", ha=ha, transform=transform, zorder=3, color="g")
        # axes.axvline(v)

    marker("|", vstats.median)
    marker("E", vstats.mean)
    marker("*", vstats.best)

    axes.text(
        0.01,
        0.95,
        vstats.label,
        zorder=2,
        backgroundcolor=(1, 1, 0, 0.2),
        verticalalignment="top",
        horizontalalignment="left",
        transform=axes.transAxes,
    )
    axes.set_yticklabels([])


def _make_fig_colorbar(logp, fig=None):
    import matplotlib as mpl
    import pylab

    # Option 1: min to min + 4
    # vmin=-max(logp); vmax=vmin+4
    # Option 1b: min to min log10(num samples)
    # vmin=-max(logp); vmax=vmin+log10(len(logp))
    # Option 2: full range of best 98%
    snllf = pylab.sort(-logp)
    vmin, vmax = snllf[0], snllf[int(0.98 * (len(snllf) - 1))]  # robust range
    # Option 3: full range
    # vmin,vmax = -max(logp),-min(logp)

    if fig is None:
        fig = pylab.gcf()
    ax = fig.axes[-1]
    cmap = mpl.cm.copper

    # Set the colormap and norm to correspond to the data for which
    # the colorbar will be used.
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    # ColorbarBase derives from ScalarMappable and puts a colorbar
    # in a specified axes, so it has everything needed for a
    # standalone colorbar.  There are many more kwargs, but the
    # following gives a basic continuous colorbar with ticks
    # and labels.
    class MinDigitsFormatter(mpl.ticker.Formatter):
        def __init__(self, low, high):
            self.delta = high - low

        def __call__(self, x, pos=None):
            # TODO: where did format_value come from?
            # it does not exist anywhere in the project.
            # return format_value(x, self.delta)
            return "{:.3G}".format(x)

    ticks = ()  # (vmin, vmax)
    formatter = MinDigitsFormatter(vmin, vmax)
    cbar = mpl.colorbar.ColorbarBase(ax, cmap=cmap, norm=norm, ticks=ticks, format=formatter, orientation="vertical")
    # cb.set_ticks(ticks)
    # cb.set_ticklabels(labels)
    # cb.set_label('negative log likelihood')

    cbar_box = ax.get_position().bounds
    fig.text(cbar_box[0], cbar_box[1], "{:.3G}".format(vmin), va="top")
    fig.text(cbar_box[0], cbar_box[1] + cbar_box[3], "{:.3G}".format(vmax), va="bottom")

    return cbar


def _make_logp_histogram(values, logp, nbins, ci, weights, cbar, axes):
    from numpy import ones_like, searchsorted, linspace, cumsum, diff, unique, argsort, array, hstack, exp

    if weights is None:
        weights = ones_like(logp)
    # TODO: values are being sorted to collect stats and again to plot
    idx = argsort(values)
    values, weights, logp = values[idx], weights[idx], logp[idx]
    # open('/tmp/out','a').write("ci=%s, range=%s\n"
    #                           % (ci,(min(values),max(values))))
    edges = linspace(ci[0], ci[1], nbins + 1)
    idx = searchsorted(values[1:-1], edges)
    # weightsum = cumsum(weights)
    # heights = diff(weightsum[idx])/weightsum[-1]  # normalized weights

    edgecolors = None
    cmap = cbar.cmap
    cmap_edges = linspace(0, 1, cmap.N + 1)[1:-1]
    bins = []  # marginalized maximum likelihood
    for s, e, xlo, xhi in zip(idx[:-1], idx[1:], edges[:-1], edges[1:]):
        if s == e:
            continue
        # parameter interval endpoints
        x = array([xlo, xhi], "d")
        # -logp values within interval, with sort index from low to high
        pv = -logp[s:e]
        pidx = argsort(pv)
        pv = pv[pidx]
        # weights for samples within interval, sorted
        pw = weights[s:e][pidx]
        # vertical colorbar top edges is the cumulative sum of the weights
        y_top = cumsum(pw)

        # For debugging compare with one rectangle per sample
        if False:
            import matplotlib as mpl

            cmap = mpl.cm.flag
            edgecolors = "k"
            xmid = (xlo + xhi) / 2
            x = [xlo, xmid]
            y = hstack((0, y_top))
            z = pv[:, None]
            pylab.pcolormesh(x, y, z, norm=cbar.norm, cmap=cmap)
            x = [xmid, xhi]

        # Possibly millions of samples, so group those which have the
        # same colour instead of drawing each as its own rectangle.
        #
        # Norm the values then look up the colormap edges in the sorted
        # normed negative log probabilities.  Drop duplicates, which
        # represent zero-width bars. Assign the value for each interval
        # according to the value at the change point.
        #
        # The indexing logic is very ugly. The searchsorted() function
        # returns 0 if before the first or N if after the last, so the
        # end points of the range are dropped so that there is and implicit
        # [-inf, ... interior points ..., inf] range. Similarly, colours
        # below vmin go to vmin and above vmax go to vmax, so drop those
        # end points as well. Then put the end points back on in the
        # found indices [0, ... interior edges ..., N-1]. Use the value
        # at the end of the boundary to colour the section.
        # Something is not quite right: with this algorithm the first
        # block appears to always be one element long and is often the
        # same colour as the next block. This is only visible if edges
        # are drawn so ignore it for now.
        change_point = searchsorted(cbar.norm(pv[1:-1]), cmap_edges)
        tops = unique(hstack((change_point, len(pv) - 1)))
        y = hstack((0, y_top[tops]))
        z = pv[tops][:, None]
        axes.pcolormesh(x, y, z, norm=cbar.norm, cmap=cmap, edgecolors=edgecolors)

        # centerpoint, histogram height, maximum likelihood for each bin
        bins.append(((xlo + xhi) / 2, y_top[-1], exp(cbar.norm.vmin - pv[0])))
    # Check for broken distribution
    if not bins:
        return
    centers, height, maxlikelihood = array(bins).T
    # Normalize maximum likelihood plot so it contains the same area as the
    # histogram, unless it is really spikey, in which case make sure it has
    # about the same height as the histogram.
    maxlikelihood *= np.sum(height) / np.sum(maxlikelihood)
    hist_peak = np.max(height)
    ml_peak = np.max(maxlikelihood)
    if ml_peak > hist_peak * 1.3:
        maxlikelihood *= hist_peak * 1.3 / ml_peak
    axes.plot(centers, maxlikelihood, "-g")

    ## plot marginal gaussian approximation along with histogram
    # def G(x, mean, std):
    #    return np.exp(-((x-mean)/std)**2/2)/np.sqrt(2*np.pi*std**2)
    ## TODO: use weighted average for standard deviation
    # mean, std = np.average(values, weights=weights), np.std(values, ddof=1)
    # pdf = G(centers, mean, std)
    # pylab.plot(centers, pdf*np.sum(height)/np.sum(pdf), '-b')


def _make_var_histogram(values, logp, nbins, ci, weights):
    # Produce a histogram
    hist, bins = np.histogram(
        values,
        bins=nbins,
        range=ci,
        # new=True,
        density=True,
        weights=weights,
    )

    # Find the max likelihood for values in each bin
    edges = np.searchsorted(values, bins)
    histbest = [np.max(logp[edges[i] : edges[i + 1]]) if edges[i] < edges[i + 1] else -np.inf for i in range(nbins)]

    # scale to marginalized probability with peak the same height as hist
    histbest = np.exp(np.asarray(histbest) - max(logp)) * np.max(hist)

    import pylab

    # Plot the histogram
    pylab.bar(bins[:-1], hist, width=bins[1] - bins[0])

    # Plot the kernel density estimate
    # density = KDE1D(values)
    # x = linspace(bins[0],bins[-1],100)
    # pylab.plot(x, density(x), '-k')

    # Plot the marginal maximum likelihood
    centers = (bins[:-1] + bins[1:]) / 2
    pylab.plot(centers, histbest, "-g")



from ...parameter import Parameter
from .. import stop
from ..solver import Minimizer
from .de import DifferentialEvolution, best1


class Function(object):
    def __init__(self, f, ndim=None, po=None, bounds=None, args=()):
        if bounds is not None and po is not None:
            self.parameters = [Parameter(value=v, bounds=b) for v, b in zip(po, bounds)]
        elif bounds is not None:
            self.parameters = [Parameter(b) for b in bounds]
        elif po is not None:
            self.parameters = [Parameter(v) for v in po]
        elif ndim is not None:
            self.parameters = [Parameter() for _ in range(ndim)]
        else:
            raise TypeError("Need ndim, po or bounds to get problem dimension")
        if (
            (ndim is not None and ndim != len(self.parameters))
            or (po is not None and len(po) != len(self.parameters))
            or (bounds is not None and len(bounds) != len(self.parameters))
        ):
            raise ValueError("Inconsistent dimensions for ndim, po and bounds")
        if po is None:
            po = [p.start_value() for p in self.parameters]

        self.f = f
        self.bounds = bounds
        self.po = po
        self.args = args

    def guess(self):
        if self.po is not None:
            return self.po
        else:
            return [p.start_value() for p in self.parameters]

    def __call__(self, p):
        return self.f(p, *self.args)


def diffev(
    func,
    x0=None,
    npop=10,
    args=(),
    bounds=None,
    ftol=5e-3,
    gtol=None,
    maxiter=None,
    maxfun=None,
    CR=0.9,
    F=0.8,
    full_output=0,
    disp=1,
    retall=0,
    callback=None,
):
    """\
Minimize a function using differential evolution.

Inputs::

    *func* -- the callable function to be minimized.
    *x0* -- the initial guess, or none for entirely random population.
    *npop* -- points per dimension.  Population size is npop*nparameters.

Additional Inputs::

    *args* -- extra arguments for func.
    *bounds* -- list of bounds (min,max), one pair for each parameter.
    *ftol* -- acceptable relative error in func(xopt) for convergence.
    *gtol* -- maximum number of iterations to run without improvement.
    *maxiter* -- the maximum number of iterations to perform.
    *maxfun* -- the maximum number of function evaluations.
    *CR* -- the probability of cross-parameter mutations
    *F* -- multiplier for impact of mutations on trial solution.
    *full_output* -- non-zero if fval and warnflag outputs are desired.
    *disp* -- non-zero to print convergence messages.
    *retall* -- non-zero to return list of solutions at each iteration.
    *callback* -- function(xk) to call after each iteration.

Returns:: (xopt, {fopt, iter, funcalls, status}, [allvecs])

    *xopt* -- the best point
    *fopt* -- value of function at the best point: fopt = func(xopt)
    *iter* -- number of iterations
    *funcalls* -- number of function calls
    *status* -- termination status
        0 : Function converged within tolerance.
        1 : Maximum number of function evaluations.
        2 : Maximum number of iterations.
    *allvecs* -- a list of solutions at each iteration
"""

    problem = Function(f=func, args=(), po=x0, bounds=None)
    strategy = DifferentialEvolution(CR=CR, F=F, npop=npop, mutate=best1)
    ndim = len(x0)

    # Determine success and failure conditions
    if gtol:  # Improvement window specified
        # look for ftol improvement over gtol generations
        success = stop.Df(tol=ftol, n=gtol, scaled=False)
    else:
        # look for f < ftol.
        success = stop.Cf(tol=ftol, scaled=False)
    if maxiter is None:
        maxiter = ndim * 100
    if maxfun is None:
        maxfun = npop * maxiter
    failure = stop.Calls(maxfun) | stop.Steps(maxiter)

    monitors = []
    # if callback is not None:
    #    monitors.append(CallbackMonitor(callback))
    # if retall:
    #    population_monitor = StepMonitor('population_values')
    #    monitors.append(population_monitor)
    minimize = Minimizer(problem=problem, strategy=strategy, monitors=monitors, success=success, failure=failure)

    # Preserve history for output (must be after solver.reset())
    hist = minimize.history
    hist.requires(point=1, value=1, calls=1, step=1)

    # Run the solver
    minimize()

    # Generate return values (must be after call to minimize)
    if hist.calls[0] > maxfun:
        status = 1
        msg = "Warning: Maximum number of function evaluations exceeded."
    elif hist.step[0] > maxiter:
        status = 2
        msg = "Warning: Maximum number of iterations exceeded."
    else:
        status = 0
        msg = """Optimization terminated successfully.
    Best point: %s
    Best value: %g
    Iterations: %d
    Function evaluations: %d""" % (hist.point[0], hist.value[0], hist.step[0], hist.calls[0])
    if disp:
        print(msg)
    if not full_output and not retall:
        ret = hist.point[0]
    elif full_output:
        ret = (hist.point[0], hist.value[0], hist.step[0], hist.calls[0], status)
        if retall:
            raise NotImplementedError("retall not implemented")
            # ret += (population_monitor.population_points,)
    else:
        raise NotImplementedError("retall not implemented")
        # ret = hist.point[0], population_monitor.population_points
    return ret



import numpy as np


def plot(draw, nlines=150, control_var=None):
    """
    Plot parallel coordinates from a draw from the distribution.

    *draw* is the draw from the sample

    *nlines* is the number of lines to draw

    If *control_var* is provided, then the best value from each bin in the
    histogram for that variable will be chosen as a line to draw on the
    coordination plot. This will will produce a roughly equally spaced set of
    coordination points for that variable. If *control_var* is None, then
    lines will be picked at random.
    """
    data = draw.points
    labels = draw.labels
    logp = draw.logp

    # TODO: confirm that plot is called with 95% interval draw
    # Note: the 95% interval draw is based on the range of the data?
    # Drop points outside the 95% interval of chisq
    # index = np.argsort(logp)
    # npoints = data.shape[0]
    # maxn = int(npoints*0.95)
    # index = index[:maxn]
    # logp = logp[index]
    # data = data[index]

    # TODO: order data by strength of correlation
    # Some options:
    #    Pearson: only works for linear
    #    Kendal, Spearman: only works for monotonic
    #    distance (Szekely, 2005) doi:10.1214/009053607000000505
    #    HGG (Heller, Heller, Gorfine, 2013) doi:10.1093/biomet/ass070
    # estimate correlations
    # corr = matrix_corr(data)
    # order = list(range(data.shape[1]))
    # data = data[order]

    if control_var is None:
        index = np.random.choice(data.shape[0], size=nlines, replace=False)
    else:
        index = best_in_bin(data[:, control_var], -logp, bins=nlines)
    # color_value, color_label = data[index, var], labels[var]
    color_value, color_label = -logp[index], "-log(P)"
    parallel_coordinates(data[index], labels=labels, value=color_value, value_label=color_label)


def best_in_bin(x, value, bins=50, range=None, keep_empty=False):
    """
    Find the index of the minimum *value* in each bin for the data in *x*.

    *x* are the coordinates to be binned.

    *value* are the objective to be minimized in each bin, such as chisq.
    There is one value for each x coordinate.

    *bins* are the number of bins within *range* or an array of bin edges
    if the bins are not uniform.

    *range* is the data range for the bins. The default is (x.min(), x.max()).

    *keep_empty* is True if empty bins will return an index of -1.  When
    False (the default), empty bins are removed.

    Returns indices of the elements which are the minimum within each bin.
    This list may be shorter than the number of bins if *keep_empty* is False.

    The algorithm works by assigning a bin number to each point then adding
    an offset in [0, 1) according to the scaled *value*.  These point values
    are then sorted, and searched by bin number.  The returned index will
    correspond to the first value in each bin, and therefore, the best value
    in that bin.  From this the index into the original list can be returned.
    """

    # Find bin number for each value.
    if isinstance(bins, int):
        # Find intervals for x and use integer division to assign bin number.
        # Don't worry that this might create bin numbers such as -3 or nbins+5
        # since these will be ignored during bin lookup.
        xmin, xmax = range if range is not None else (np.min(x), np.max(x))
        dx = (xmax - xmin) / bins
        value_in_bin = (x - xmin) // dx if dx != 0.0 else np.full_like(x, bins // 2)
        nbins = bins
    else:
        # Lookup x in bin edges. searchsorted returns index 0 for elements
        # before the first edge so we need to put an extra start edge and
        # subtract one from the index so that bin number is -1 before first.
        value_in_bin = np.searchsorted(np.hstack((-np.inf, bins)), x) - 1
        nbins = len(bins) - 1

    # Increment bin number offset using scaled value.  Limit the scaled
    # value to 0.99 rather 1.0 so that the worst value isn't 1.0, which
    # will be the best value in the next bin.
    value_in_bin += scale(value) * 0.99

    # Sort values, preserving original indices
    sort_index = np.argsort(value_in_bin)
    sorted_value_in_bin = value_in_bin[sort_index]

    # Lookup first point in each bin, which will now by the min chisq in bin.
    bin_numbers = np.arange(nbins)
    min_index = np.searchsorted(sorted_value_in_bin, bin_numbers)

    # Find the point number for the best points in the original list.
    index = sort_index[min_index]

    # Indentify empty bins. This will occur if base value returned by
    # search sorted does not match the bin number.
    empty = np.floor(sorted_value_in_bin[min_index]) != bin_numbers

    # Process empty bins.
    if keep_empty:
        index[empty] = -1
    else:
        index = index[~empty]

    return index


def parallel_coordinates(data, labels=None, value=None, value_label=""):
    """
    Produce a parallel coordinates plot.

    *data* is a set of points to draw, one row per point.

    *labels* is the label to assign to the dimensions, one per column in *data*.

    *value* is an optional value to use to color the different lines. This
    could be the value of the control variable, or some additional dimension
    such as overall quality for the individual points.

    *value_label* is the label to put on the colorbar for the line colors.
    """
    import matplotlib.pyplot as plt
    from matplotlib.collections import LineCollection

    ndim = data.shape[1]
    if labels is None:
        labels = ["p%d" % k for k in range(ndim)]

    x = np.arange(ndim)
    data = scale(data, axis=0)

    # We need to set the plot limits, they will not autoscale
    fig, ax = plt.gcf(), plt.gca()
    ax.set_xlim(0, ndim - 1)
    ax.set_ylim(0, 1)
    plt.xticks(x, labels, rotation=30)
    lines = LineCollection([np.column_stack([x, y]) for y in data], linewidths=1, linestyles="solid")
    ax.add_collection(lines)
    ax.set_yticklabels([])

    if value is not None:
        lines.set_array(value)
        cbar = fig.colorbar(lines)
        cbar.set_label(value_label)
        plt.sci(lines)


def scale(x, axis=None):
    """
    Returns *x* with values scaled to [0, 1].

    If *axis* is not None, then scale values within each axis independently,
    otherwise use the min/max value across all dimensions.
    """
    low = x.min(axis=axis, keepdims=True)
    high = x.max(axis=axis, keepdims=True)
    dx = high - low
    dx[dx == 0.0] = 1.0
    scaled = (x - low) / dx
    return scaled



"""
ACR upper percentiles critical value for test of single multivariate
normal outlier.

From the method given by Wilks (1963) and approaching to a F distribution
function by the Yang and Lee (1987) formulation, we compute the critical
value of the maximum squared Mahalanobis distance to detect outliers from
a normal multivariate sample.

We can generate all the critical values of the maximum squared Mahalanobis
distance presented on the Table XXXII of by Barnett and Lewis (1978) and
Table A.6 of Rencher (2002). Also with any given significance level (alpha).

Example::

    >>> print("%.4f"%ACR(3, 25, 0.01))
    13.1753

Created by::

    A. Trujillo-Ortiz, R. Hernandez-Walls, A. Castro-Perez and K. Barba-Rojo
    Facultad de Ciencias Marinas
    Universidad Autonoma de Baja California
    Apdo. Postal 453
    Ensenada, Baja California
    Mexico.
    atrujo@uabc.mx

Copyright. August 20, 2006.

To cite this file, this would be an appropriate format::

    Trujillo-Ortiz, A., R. Hernandez-Walls, A. Castro-Perez and K. Barba-Rojo.
    (2006). *ACR:Upper percentiles critical value for test of single
    multivariate  normal outlier.* A MATLAB file. [WWW document].  URL
    http://www.mathworks.com/matlabcentral/fileexchange/loadFile.do?objectId=12161

The function's name is given in honour of Dr. Alvin C. Rencher for his
invaluable contribution to multivariate statistics with his text 'Methods of
Multivariate Analysis'.

References:

[1] Barnett, V. and Lewis, T. (1978), Outliers on Statistical Data.
     New-York:John Wiley & Sons.
[2] Rencher, A. C. (2002), Methods of Multivariate Analysis. 2nd. ed.
     New-Jersey:John Wiley & Sons. Chapter 13 (pp. 408-450).
[3] Wilks, S. S. (1963), Multivariate Statistical Outliers. Sankhya,
     Series A, 25: 407-426.
[4] Yang, S. S. and Lee, Y. (1987), Identification of a Multivariate
     Outlier. Presented at the Annual  Meeting of the American
     Statistical Association, San Francisco, August 1987.
"""

__all__ = ["ACR"]

from scipy.stats import f

finv = f.ppf


def ACR(p, n, alpha=0.05):
    """
    Return critical value for test of single multivariate normal outlier
    using the Mahalanobis distance metric.

    *p* is the number of independent variables,
    *n* is the number of samples, and
    *alpha* is the significance level cutoff (default=0.05).
    """

    if alpha <= 0 or alpha >= 1:
        raise ValueError("significance level must be between 0 and 1")

    a = alpha
    # F distribution critical value with p and n-p-1 degrees of freedom
    # using the Bonferroni correction.
    fc = finv(1 - a / n, p, n - p - 1)
    result = (p * (n - 1) ** 2 * fc) / (n * (n - p - 1) + (n * p * fc))
    # = ((-1*((1/(1+(fc*p/(n-p-1))))-1))*((n-1)^2))/n;

    return result


def test():
    assert abs(ACR(3, 25, 0.01) - 13.1753251622586) < 1e-13


if __name__ == "__main__":
    test()



"""
Convergence test statistic from Gelman and Rubin, 1992.[1]

[1] Gelman, Andrew, and Donald B. Rubin.
    "Inference from Iterative Simulation Using Multiple Sequences."
    Statistical Science 7, no. 4 (November 1, 1992): 457-72.
    https://doi.org/10.2307/2246093.
"""

__all__ = ["gelman"]

from numpy import var, mean, ones, sqrt


def gelman(sequences, portion=0.5):
    """
    Calculates the R-statistic convergence diagnostic

    For more information please refer to: Gelman, A. and D.R. Rubin, 1992.
    Inference from Iterative Simulation Using Multiple Sequences,
    Statistical Science, Volume 7, Issue 4, 457-472.
    doi:10.1214/ss/1177011136
    """

    # Find the size of the sample
    chain_len, nchains, nvar = sequences.shape
    # print sequences[:20, 0, 0]

    # Only use the last portion of the sample
    chain_len = int(chain_len * portion)
    sequences = sequences[-chain_len:]

    if chain_len < 2:
        # Set the R-statistic to a large value
        r_stat = -2 * ones(nvar)
    else:
        # Step 1: Determine the sequence means
        mean_seq = mean(sequences, axis=0)

        # Step 1: Determine the variance between the sequence means
        b = chain_len * var(mean_seq, axis=0, ddof=1)

        # Step 2: Compute the variance of the various sequences
        var_seq = var(sequences, axis=0, ddof=1)

        # Step 2: Calculate the average of the within sequence variances
        w = mean(var_seq, axis=0)

        # Step 3: Estimate the target mean
        # mu = mean(mean_seq)

        # Step 4: Estimate the target variance (Eq. 3)
        sigma2 = ((chain_len - 1) / chain_len) * w + (1 / chain_len) * b

        # TODO: the second term, -(N-1)/(K N), doesn't appear in [1]
        # Step 5: Compute the R-statistic
        r_stat = sqrt((nchains + 1) / nchains * sigma2 / w - (chain_len - 1) / nchains / chain_len)
        # par=2
        # print chain_len,b[par],var_seq[...,par],w[par],r_stat[par]

    return r_stat


def test():
    from numpy import reshape, arange, transpose
    from numpy.linalg import norm

    # Targe values computed from octave:
    #    format long
    #    s = reshape([1:15*6*7],[15,6,7]);
    #    r = gelman(s,struct('n',6,'seq',7))
    s = reshape(arange(1.0, 15 * 6 * 7 + 1) ** -2, (15, 6, 7), order="F")
    s = transpose(s, [0, 2, 1])
    target = [
        1.06169861367116,
        2.75325774624905,
        4.46256647696399,
        6.12792266170178,
        7.74538715553575,
        9.31276519155232,
    ]
    r = gelman(s, portion=1)
    # print r
    # print "target", array(target), "\nactual", r
    assert norm(r - target) < 1e-14
    r = gelman(s, portion=0.1)
    assert norm(r - [-2, -2, -2, -2, -2, -2]) == 0


if __name__ == "__main__":
    test()



# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author: James Krycka

"""
This module contains a custom About Dialog class and associated text strings
used for informational display purposes.  Note that the product version is
maintained in the version.py file and therefore is imported here.
"""

import os
import wx

try:
    from agw.hyperlink import HyperLinkCtrl
except ImportError:  # if it's not there, try the older location.
    from wx.lib.agw.hyperlink import HyperLinkCtrl

from wx.lib.wordwrap import wordwrap

from .. import __version__ as APP_VERSION
from .utilities import resource

# Resource files.
PROG_ICON = "bumps.ico"

# Text strings used in About Dialog boxes and for other project identification
# purposes.
#
# Note that paragraphs intended to be processed by wordwrap are formatted as
# one string without newline characters.
APP_NAME = "Bumps"

APP_TITLE = "Bumps: curve fitter with uncertainty estimation"

APP_COPYRIGHT = "(C) 2011 University of Maryland"

APP_DESCRIPTION = """\
The Bumps uncertainty modeler provides an interactive user interface for \
modelling an experiment and fitting it to data.\
varying free form layers.
"""

APP_LICENSE = """\
Permission is hereby granted, free of charge, to any person obtaining a copy \
of this software and associated documentation files (the "Software"), to deal \
in the Software without restriction, including without limitation the rights \
to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell \
copies of the Software, and to permit persons to whom the Software is \
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in \
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR \
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, \
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE \
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER \
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, \
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN \
THE SOFTWARE.
"""

APP_CREDITS = """\
This program was developed jointly by the University of Maryland (UMD) and \
the National Institute of Standards and Technology (NIST) as part of the \
Distributed Data Analysis of Neutron Scattering Experiments (DANSE) project \
funded by the US National Science Foundation under grant DMR-0520547. \
Principal contributors:

Paul Kienzle, NIST
    - Application concept and design
    - Application API and Command Line Interface development
    - Optimizers and uncertainty analysis
    - Documentation

Ismet Sahin, UMD
    - Optimizers

James Krycka, UMD
    - Graphical User Interface design and development
    - Window's installer

Nikunj Patel, UMD
    - Graphical User Interface design and development

Christopher Metting, UMD
    - Artwork
"""

APP_PROJECT_URL = "http://github.com/bumps"
APP_PROJECT_TAG = "Bumps home page"

APP_TUTORIAL_URL = "http://pypi.org/docs/bumps"
APP_TUTORIAL_TAG = "Bumps documentation"
APP_TUTORIAL = """\
For the Bumps User's Guide and Reference Manual, please visit:\
"""

# ==============================================================================


class AboutDialog(wx.Dialog):
    """
    This class creates a pop-up About Dialog box with several display options.
    """

    def __init__(
        self,
        parent=None,
        id=wx.ID_ANY,
        title="About",
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.DEFAULT_DIALOG_STYLE,
        show_name=True,
        show_notice=True,
        show_link=True,
        show_link_docs=False,
        info="...",
    ):
        wx.Dialog.__init__(self, parent, id, title, pos, size, style)

        # Display the application's icon in the title bar.
        icon = wx.Icon(resource(PROG_ICON), wx.BITMAP_TYPE_ICO)
        self.SetIcon(icon)

        # Set the font for this window and all child windows (widgets) from the
        # parent window, or from the system defaults if no parent is given.
        # A dialog box does not inherit font info from its parent, so we will
        # explicitly get it from the parent and apply it to the dialog box.
        if parent is not None:
            font = parent.GetFont()
            self.SetFont(font)

        # Display program name and version.
        if show_name:
            prog = wx.StaticText(self, wx.ID_ANY, label=(APP_NAME + " " + APP_VERSION))
            font = prog.GetFont()
            font.SetPointSize(font.GetPointSize() + 1)
            font.SetWeight(wx.BOLD)
            prog.SetFont(font)

        # Display copyright notice.
        if show_notice:
            copyright = wx.StaticText(self, wx.ID_ANY, label=APP_COPYRIGHT)

        # Display hyperlink to the home page and/or doc page.
        if show_link:
            hyper1 = HyperLinkCtrl(self, wx.ID_ANY, label=APP_PROJECT_TAG, URL=APP_PROJECT_URL)
        if show_link_docs:
            hyper2 = HyperLinkCtrl(self, wx.ID_ANY, label=APP_TUTORIAL_TAG, URL=APP_TUTORIAL_URL)

        # Display the body of text for this about dialog box.
        info = wx.StaticText(self, wx.ID_ANY, label=wordwrap(info, 530, wx.ClientDC(self)))
        # Create the OK button control.
        ok_button = wx.Button(self, wx.ID_OK, "OK")
        ok_button.SetDefault()

        # Use a vertical box sizer to manage the widget layout..
        sizer = wx.BoxSizer(wx.VERTICAL)
        if show_name:
            sizer.Add(prog, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, border=10)
        if show_notice:
            sizer.Add(copyright, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, border=10)
        sizer.Add(info, 0, wx.ALL, border=10)
        if show_link:
            sizer.Add(hyper1, 0, wx.ALL, border=10)
        if show_link_docs:
            sizer.Add(hyper2, 0, wx.ALL, border=10)
        sizer.Add(ok_button, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, border=10)

        # Finalize the sizer and establish the dimensions of the dialog box.
        self.SetSizer(sizer)
        sizer.Fit(self)



"""
Interactive console widget support.

Defines NumpyConsole class.

TODO: Fix cut/paste for multiline commands
TODO: Trigger change notification when numpy array has changed
"""

import wx, wx.py


def shapestr(v):
    """Return shape string for numeric variables suitable for printing"""
    try:
        shape = v.shape
    except AttributeError:
        return "scalar"
    else:
        return "array " + "x".join([str(i) for i in shape])


class NumpyConsole(wx.py.shell.ShellFrame):
    """
    NumpyConsole defines an interactive console which is aware of all the
    numerical variables in the local name space.  When variables are added
    or removed, it signals that the set of variables has changed.

    This is intended to be used as an embedded console in an interactive
    application for which the user can define and manipulate numerical
    types and automatically show a list of variables, e.g., available for
    plotting.

    If you subclass and replace self.init_code, be sure to do so _before_ calling
    the superclass __init__.
    """

    # code to define the initial namespace
    init_code = """
from pylab import *
"""
    introText = """
Welcome to the numpy console!

Example:

x = linspace(0,1,300)
y = sin(2*pi*x*3)
vars()
plot(x,y)
"""

    def __init__(self, *args, **kwargs):
        # Start interpreter and monitor statement execution
        msg = kwargs.pop("introText", self.introText)
        wx.py.shell.ShellFrame.__init__(self, *args, **kwargs)

        # Print welcome message.
        # TODO: replace this when (if?) ShellFrame allows introText=msg
        print(msg, file=self.shell)
        self.shell.prompt()

        # Initialize the interpreter namespace with useful commands
        self.shell.interp.runcode(compile(self.init_code, "__main__", "exec"))

        # steal draw_if_interactive
        import pylab
        from matplotlib._pylab_helpers import Gcf
        from matplotlib import pyplot

        self._dirty = set()

        def draw_if_interactive():
            # print "calling draw_if_interactive with",Gcf.get_active()
            self._dirty.add(Gcf.get_active())

        pyplot.draw_if_interactive = draw_if_interactive

        # add vars command to the interpreter
        self.shell.interp.locals["vars"] = self._print_vars

        # ignore the variables defined by numpy
        self.ignore = set(self.shell.interp.locals.keys())
        self._existing = {}  # No new variables recorded yet

        # remember which variables are current so we can detect changes
        wx.py.dispatcher.connect(receiver=self._onPush, signal="Interpreter.push")

    def filter(self, key, value):
        """
        Return True if var should be listed in the available variables.
        """
        return key not in self.ignore

    # Dictionary interface
    def items(self):
        """
        Return the list of key,value pairs for all locals not ignored.
        """
        locals = self.shell.interp.locals
        for k, v in locals.items():
            if self.filter(k, v):
                yield k, v

    def update(self, *args, **kw):
        """
        Update a set of variables from a dictionary.
        """
        self.shell.interp.locals.update(*args, **kw)
        self._existing.update(*args, **kw)

    def __setitem__(self, var, val):
        """
        Define or replace a variable in the interpreter.
        """
        self.shell.interp.locals[var] = val
        self._existing[var] = val

    def __getitem__(self, var):
        """
        Retrieve a variable from the interpreter.
        """
        return self.shell.interp.locals[var]

    def __delitem__(self, var):
        """
        Delete a variable from the interpreter.
        """
        del self.shell.interp.locals[var]
        try:
            del self._existing[var]
        except KeyError:
            pass

    # Stream interface
    def write(self, msg):
        """
        Support 'print >>console, blah' for putting output on console.

        TODO: Maybe redirect stdout to console if console is open?
        """
        self.shell.write(self, msg)

    # ==== Internal messages ====
    def OnChanged(self, added=[], changed=[], removed=[]):
        """
        Override this method to perform your changed operation.

        Note that we cannot detect changes within a variable without considerable
        effort: we would need to keep a deep copy of the original value, and
        use a deep comparison to see if it has changed.
        """
        for var in added:
            print("added", var, file=self.shell)
        for var in changed:
            print("changed", var, file=self.shell)
        for var in removed:
            print("deleted", var, file=self.shell)
        print("override the OnChanged message to update your application state", file=self.shell)

    def _print_vars(self):
        """
        Print the available numeric variables and their shapes.

        This is a command available to the user as vars().
        """
        locals = self.shell.interp.locals
        for k, v in self.items():
            print(k, shapestr(v), file=self.shell)

    def _onPush(self, **kw):
        """On command execution, detect if variable list has changed."""
        # Note: checking for modify is too hard ... build it into the types?
        # print >>self.shell, "checking for add/delete..."
        # Update graphs if changed
        if self._dirty:
            import pylab
            from matplotlib._pylab_helpers import Gcf

            # print "figs",Gcf.figs
            # print "dirty",self._dirty
            for fig in self._dirty:
                # print fig, Gcf.figs.values(),fig in Gcf.figs.values()
                if fig and fig in Gcf.figs.values():
                    # print "drawing"
                    fig.canvas.draw()
            pylab.show()
            self._dirty.clear()

        items = dict(list(self.items()))
        oldkeys = set(self._existing.keys())
        newkeys = set(items.keys())
        added = newkeys - oldkeys
        removed = oldkeys - newkeys
        changed = set(k for k in (oldkeys & newkeys) if items[k] is not self._existing[k])
        if added or changed or removed:
            self.OnChanged(added=added, changed=changed, removed=removed)
        self._existing = items


def demo():
    """Example use of the console."""
    import numpy as np

    app = wx.App(redirect=False)
    ignored = {"f": lambda x: 3 + x}
    console = NumpyConsole(locals=ignored)
    console.update({"x": np.array([[42, 15], [-10, 12]]), "z": 42.0})
    console.Show(True)
    app.MainLoop()


if __name__ == "__main__":
    demo()



r"""
Least squares error analysis.

Given a data set with gaussian uncertainty on the points, and a model which
is differentiable at the minimum, the parameter uncertainty can be estimated
from the covariance matrix at the minimum.  The model and data are wrapped in
a problem object, which must define the following methods:

    ============ ============================================
    getp()       get the current value of the model
    setp(p)      set a new value in the model
    nllf(p)      negative log likelihood function
    residuals(p) residuals around its current value
    bounds()     get the bounds on the parameter p [optional]
    ============ ============================================

:func:`jacobian` computes the Jacobian matrix $J$ using numerical
differentiation on residuals. Derivatives are computed using the center
point formula, with two evaluations per dimension.  If the problem has
analytic derivatives with respect to the fitting parameters available,
then these should be used to compute the Jacobian instead.

:func:`hessian` computes the Hessian matrix $H$ using numerical
differentiation on nllf.

:func:`jacobian_cov` takes the Jacobian and computes the covariance matrix $C$.
:func:`hessian_cov` takes the Hessian and computes the covariance matrix $C$.

:func:`corr` uses the off-diagonal elements of $C$ to compute correlation
coefficients $R^2_{ij}$ between the parameters.

:func:`stderr` computes the uncertain $\sigma_i$ from covariance matrix $C$,
assuming that the $C_\text{diag}$ contains $\sigma_i^2$, which should be
the case for functions which are approximately linear near the minimum.

:func:`max_correlation` takes $R^2$ and returns the maximum correlation.

The user should be shown the uncertainty $\sigma_i$ for each parameter,
and if there are strong parameter correlations (e.g., $R^2_\text{max} > 0.2$),
the correlation matrix as well.

The bounds method for the problem is optional, and is used only to determine
the step size needed for the numerical derivative.  If bounds are not present
and finite, the current value for the parameter is used as a basis to
estimate step size.

"""

import numpy as np


def gradient(problem, p=None, step=None):
    r = problem.residuals()
    J = jacobian(problem, p=p, step=step)
    return np.dot(J.T, r)


# TODO: restructure lsqerror to use mapper for evaluating multiple f
# doesn't work for jacobian since mapper returns nllf; would need to
# expand mapper to implement a variety of different functions.
def jacobian(problem, p=None, step=None):
    """
    Returns the derivative wrt the fit parameters at point p.

    Numeric derivatives are calculated based on step, where step is
    the portion of the total range for parameter j, or the portion of
    point value p_j if the range on parameter j is infinite.

    The current point is preserved.

    Note that the problem.residuals() method should not reuse memory for the
    returned value otherwise the derivative calculation (f(x+dx) - f(x))/dx
    will always be zero. The returned value need not be 1D, but it should be
    contiguous so that it can be reshaped to 1D without an extra copy. This
    will only be an issue for very large datasets.
    """
    p_init = problem.getp()
    if p is None:
        p = p_init
    p = np.asarray(p)
    bounds = getattr(problem, "bounds", lambda: None)()

    def f(p):
        problem.setp(p)
        # Return residuals as a vector even if f(x) returns a matrix otherwise
        # we cannot build a stacked Jacobian. We use reshape() rather than
        # flatten since this will avoid an unnecessary copy.
        return np.reshape(problem.residuals(), -1)

    J = _jacobian_forward(f, p, bounds, eps=step)
    problem.setp(p_init)
    return J


def _jacobian_forward(f, p, bounds, eps=None):
    n = len(p)
    # TODO: default to double precision epsilon
    step = 1e-4 if eps is None else np.sqrt(eps)

    # print("p",p,"step",step)
    h = abs(p) * step
    h[h == 0] = step
    if bounds is not None:
        h[h + p > bounds[1]] *= -1.0  # step backward if forward step is out of bounds
    ee = np.diag(h)

    fx = f(p)  # Maybe fx.copy() to protect against reuse
    J = []
    for i in range(n):
        fx_plus = f(p + ee[i, :])
        J.append((fx_plus - fx) / h[i])
    return np.vstack(J).T


def _jacobian_central(f, p, bounds, eps=None):
    n = len(p)
    # TODO: default to double precision epsilon
    step = 1e-4 if eps is None else np.sqrt(eps)

    # print("p",p,"step",step)
    h = abs(p) * step
    h[h == 0] = step
    # if bounds is not None:
    #    h[h+p>bounds[1]] *= -1.0  # step backward if forward step is out of bounds
    ee = np.diag(h)

    J = []
    for i in range(n):
        fx_minus = f(p - ee[i, :])  # Maybe fx.copy() to protect against reuse
        fx_plus = f(p + ee[i, :])
        J.append((fx_plus - fx_minus) / (2.0 * h[i]))
    return np.vstack(J).T


def hessian(problem, p=None, step=None):
    """
    Returns the derivative wrt to the fit parameters at point p.

    The current point is preserved.
    """
    p_init = problem.getp()
    if p is None:
        p = p_init
    p = np.asarray(p)
    bounds = getattr(problem, "bounds", lambda: None)()
    H = _hessian_forward(problem.nllf, p, bounds=bounds, eps=step)
    problem.setp(p_init)
    return H


def _hessian_forward(f, p, bounds, eps=None):
    # type: (Callable[[np.ndarray], float], np.ndarray, Optional[np.ndarray]) -> np.ndarray
    """
    Forward difference Hessian.
    """
    n = len(p)
    # TODO: default to double precision epsilon
    step = 1e-4 if eps is None else np.sqrt(eps)
    fx = f(p)

    # print("p",p,"step",step)
    h = abs(p) * step
    h[h == 0] = step
    if bounds is not None:
        h[h + p > bounds[1]] *= -1.0  # step backward if forward step is out of bounds
    ee = np.diag(h)

    g = np.empty(n, "d")
    for i in range(n):
        g[i] = f(p + ee[i, :])
    # print("fx",fx)
    # print("h",h, h[0])
    # print("g",g)
    H = np.empty((n, n), "d")
    for i in range(n):
        for j in range(i, n):
            fx_ij = f(p + ee[i, :] + ee[j, :])
            # print("fx_%d%d=%g"%(i,j,fx_ij))
            H[i, j] = (fx_ij - g[i] - g[j] + fx) / (h[i] * h[j])
            H[j, i] = H[i, j]
    return H


def _hessian_central(f, p, bounds, eps=None):
    # type: (Callable[[np.ndarray], float], np.ndarray, Optional[np.ndarray]) -> np.ndarray
    """
    Central difference Hessian.
    """
    n = len(p)
    # TODO: default to double precision epsilon
    step = 1e-4 if eps is None else np.sqrt(eps)
    # step = np.sqrt(step)
    fx = f(p)

    h = abs(p) * step
    h[h == 0] = step
    # TODO: handle bounds on central difference formula
    # if bounds is not None:
    #    h[h+p>bounds[1]] *= -1.0  # step backward if forward step is out of bounds
    ee = np.diag(h)

    gp = np.empty(n, "d")
    gm = np.empty(n, "d")
    for i in range(n):
        gp[i] = f(p + ee[i, :])
        gm[i] = f(p - ee[i, :])
    H = np.empty((n, n), "d")
    for i in range(n):
        for j in range(i, n):
            fp_ij = f(p + ee[i, :] + ee[j, :])
            fm_ij = f(p - ee[i, :] - ee[j, :])
            # print("fx_%d%d=%g"%(i,j,fx_ij))
            H[i, j] = (fp_ij - gp[i] - gp[j] + fm_ij - gm[i] - gm[j] + 2.0 * fx) / (2.0 * h[i] * h[j])
            H[j, i] = H[i, j]
    return H


def perturbed_hessian(H, scale=None):
    """
    **DEPRECATED** Numerical testing has shown that the perturbed Hessian
    is too aggressive with its perturbation, and it is distorting the error
    too much, so use hessian_cov(H) instead.

    Adjust Hessian matrix to be positive definite.

    Returns the adjusted Hessian and its Cholesky decomposition.
    """
    from .quasinewton import modelhess

    n = H.shape[0]
    if scale is None:
        scale = np.ones(n)
    macheps = np.finfo("d").eps
    return modelhess(n, scale, macheps, H)


def chol_stderr(L):
    """
    Return parameter uncertainty from the Cholesky decomposition of the
    Hessian matrix, as returned, e.g., from the quasi-Newton optimizer BFGS
    or as calculated from :func:`perturbed_hessian` on the output of
    :func:`hessian` applied to the cost function problem.nllf.

    Note that this calls chol_cov to compute the inverse from the Cholesky
    decomposition, so use stderr(C) if you are already computing C = chol_cov().

    **Warning:** assumes H = L@L.T (numpy default) not H = U.T@U (scipy default).
    """
    # TODO: are there numerical tricks to get the diagonal without the full inv?
    return stderr(chol_cov(L))


def chol_cov(L):
    """
    Given the cholesky decomposition of the Hessian matrix H, compute
    the covariance matrix $C = H^{-1}$

    **Warning:** assumes H = L@L.T (numpy default) not H = U.T@U (scipy default).
    """
    Linv = np.linalg.inv(L)
    return np.dot(Linv.T.conj(), Linv)


def jacobian_cov(J, tol=1e-8):
    """
    Given Jacobian J, return the covariance matrix inv(J'J).

    We provide some protection against singular matrices by setting
    singular values smaller than tolerance *tol* to the tolerance
    value.
    """
    # Find cov of f at p
    #     cov(f,p) = inv(J'J)
    # Use SVD
    #     J = U S V'
    #     J'J = (U S V')' (U S V')
    #         = V S' U' U S V'
    #         = V S S V'
    #     inv(J'J) = inv(V S S V')
    #              = inv(V') inv(S S) inv(V)
    #              = V inv (S S) V'
    u, s, vh = np.linalg.svd(J, 0)
    s[s <= tol] = tol
    JTJinv = np.dot(vh.T.conj() / s**2, vh)
    return JTJinv


def hessian_cov(H, tol=1e-15):
    """
    Given Hessian H, return the covariance matrix inv(H).

    We provide some protection against singular matrices by setting
    singular values smaller than tolerance *tol* (relative to the largest
    singular value) to zero (see np.linalg.pinv for details).
    """
    # Find cov of f at p
    #     cov(f,p) = inv(H)
    # Use SVD
    #     H = U S V'
    #     inv(H) = inv(U S V')
    #            = inv(V') inv(S S) inv(U)
    #            = V inv(S S) U'
    #     J'J = (U S V')' (U S V')
    #         = V S' U' U S V'
    #         = V S S V'
    #     inv(J'J) = inv(V S S V')
    #              = inv(V') inv(S S) inv(V)
    #              = V inv (S S) V'
    return np.linalg.pinv(H, rcond=tol, hermitian=True)


def corr(C):
    """
    Convert covariance matrix $C$ to correlation matrix $R^2$.

    Uses $R = D^{-1} C D^{-1}$ where $D$ is the square root of the diagonal
    of the covariance matrix, or the standard error of each variable.
    """
    Dinv = 1.0 / stderr(C)
    return np.dot(Dinv, np.dot(C, Dinv))


def max_correlation(Rsq):
    """
    Return the maximum correlation coefficient for any pair of variables
    in correlation matrix Rsq.
    """
    return np.max(np.tril(Rsq, k=-1))


def stderr(C):
    r"""
    Return parameter uncertainty from the covariance matrix C.

    This is just the square root of the diagonal, without any correction
    for covariance.

    If measurement uncertainty is unknown, scale the returned uncertainties
    by $\sqrt{\chi^2_N}$, where $\chi^2_N$ is the sum squared residuals
    divided by the degrees  of freedom.  This will match the uncertainty on
    the parameters to the observed scatter assuming the model is correct and
    the fit is optimal.  This will also be appropriate for weighted fits
    when the true measurement uncertainty dy_i is known up to a scaling
    constant for all y_i.

    Standard error on scipy.optimize.curve_fit always includes the chisq
    correction, whereas scipy.optimize.leastsq never does.
    """
    return np.sqrt(np.diag(C))


def demo_hessian():
    rosen = lambda x: (1.0 - x[0]) ** 2 + 105 * (x[1] - x[0] ** 2) ** 2
    p = np.array([1.0, 1.0])
    H = _hessian_forward(rosen, p, bounds=None, eps=1e-16)
    print("forward difference H", H)
    H = _hessian_central(rosen, p, bounds=None, eps=1e-16)
    print("central difference H", H)


def demo_jacobian():
    y = np.array([1.0, 2.0, 3.0])
    f = lambda x: x[0] * y + x[1]
    p = np.array([2.0, 3.0])
    J = _jacobian_forward(f, p, bounds=None, eps=1e-16)
    print("forward difference J", J)
    J = _jacobian_central(f, p, bounds=None, eps=1e-16)
    print("central difference J", J)


# https://en.wikipedia.org/wiki/Hilbert_matrix
# Note: 1-origin indices translated to 0-origin
def hilbert(n):
    """Generate ill-conditioned Hilbert matrix of size n x n"""
    return 1 / (np.arange(n)[:, None] + np.arange(n)[None, :] + 1)


# https://en.wikipedia.org/wiki/Hilbert_matrix#Properties
# Note: 1-origin indices translated to 0-origin
def hilbertinv(n):
    """Analytical inverse for ill-conditioned Hilbert matrix of size n x n"""
    Hinv = [
        [
            (-1) ** (i + j + 2) * (i + j + 1) * comb(n + i, n - j - 1) * comb(n + j, n - i - 1) * comb(i + j, i) ** 2
            for i in range(n)
        ]
        for j in range(n)
    ]
    return np.asarray(Hinv, dtype="d")


# From dheerosaur
# https://stackoverflow.com/questions/4941753/is-there-a-math-ncr-function-in-python/4941932#4941932
def comb(n, r):
    """n choose r combination function"""
    import operator as op
    from functools import reduce

    r = min(r, n - r)
    numer = reduce(op.mul, range(n, n - r, -1), 1)
    denom = reduce(op.mul, range(1, r + 1), 1)
    return numer // denom


def demo_stderr_hilbert(n=5):
    H = hilbert(n)
    C = hilbertinv(n)
    s = stderr(C)
    Hp, Lp = perturbed_hessian(H)
    Cp = chol_cov(Lp)
    sp = chol_stderr(Lp)
    Cdirect = hessian_cov(H)
    sdirect = stderr(Cdirect)
    with np.printoptions(precision=3):
        print("s ", s)
        print("sp", sp)
        print("sd", sdirect)
        print("R", corr(C))


def demo_stderr_perturbed():
    n = 5
    D = [1, 2, 3, 4, 5]
    # D = np.exp(10*np.random.rand(n)**2)
    D = [1e-3, 1e-2, 1e-1, 1, 10]

    D = np.asarray(D)
    L = np.tril(np.random.rand(n, n))
    np.fill_diagonal(L, D)
    H = np.dot(L, L.T)
    Hp, Lp = perturbed_hessian(H)
    C = chol_cov(Lp)
    s = chol_stderr(Lp)

    from scipy.linalg import inv

    Ldirect = np.linalg.cholesky(H)
    Cdirect = inv(H)
    Cp = inv(Hp)

    sdirect = np.sqrt(np.diag(Cdirect))
    sp = np.sqrt(np.diag(Cp))
    sdirect_chol = chol_stderr(Ldirect)

    parts = dict(
        L_original=L,
        L_direct=Ldirect,
        L_perturbed=Lp,
        # H=H,
        # H_perturbed=Hp,
        # C_direct=Cdirect,
        # C_from_Hp=Cp,
        # C_from_Lp=C,
    )
    with np.printoptions(precision=3):
        print("%20s" % ("perturbation"), hp[0, 0] - h[0, 0])
        for k, v in parts.items():
            print("%20s" % (k + " diag"), np.diag(v))
        # print("eigc", list(sorted(np.linalg.eigvals(c))))
        # print("eigcp", list(sorted(np.linalg.eigvals(cp))))
        # print("eigh", list(sorted(1/np.linalg.eigvals(h))))
        # print("eighp", list(sorted(1/np.linalg.eigvals(hp))))
        print("h cond     ", np.linalg.cond(h))
        print("rel err dc ", abs((c - cdirect) / cdirect).max())
        print("de         ", sp - s)
        print("s direct   ", sdirect)
        print("s chol     ", sdirect_chol)
        print("s perturbed", sp)
        print("s          ", s)
        print("rel err ds ", abs((s - sdirect) / sdirect).max())
        print("unperturbed ds", abs((sdirect_chol - sdirect) / sdirect).max())


if __name__ == "__main__":
    # demo_hessian()
    # demo_jacobian()
    # demo_stderr_perturbed()
    demo_stderr_hilbert(10)



# Identify the resource files for the bumps GUI
from ..resfiles import Resources

resources = Resources(
    package=__name__,
    patterns=("*.png", "*.jpg", "*.ico", "*.wav"),
    datadir="bumps-data",
    check_file="reload.png",
    env="BUMPS_DATA",
)
del Resources



"""
Redirect calls to open, etc. to a virtual file system.

Use this to mount a zip file as a file system, and then all subsequent calls
to chdir, open, etc. will reference files in the zip file instead of the disk.

This will only work for packages which do all their I/O in python, and not
those which use direct calls to the C library, so for example, it will not
work with h5py. XML parsing from a zip file is also unlikely to work since
expat uses the C library directly for parsing.

Usage::

    # Do this before importing any other modules!  It sets up hooks for
    # redirecting filesystem access even if the module imports the symbol
    # directly as "from os import getcwd".
    import vfs
    vfs.vfs_init()
    ...
    with vfs.ZipFS('data.zip'):
        data = np.loadtxt('file1.dat')

Filesystems available:

* `RealFS` - uses the builtin python functions to access the O/S.

* `ZipFS` - opens a zip file as the filesystem root.

Calls redirected::

    __builtin__.open (python 2 only)
    builtins.open (python 2 and 3)
    io.open (python 2 and 3)
    os, nt, posix:
        chdir, getcwd, listdir
    os.path, ntpath, posixpath:
        exists, isfile, isir, abspath, realpath

You can also use the file systems directly without using the `vfs_init`
hook or the with statement.  Just call `fs.chdir`, etc. on the file system
object.

*file* in python 2.x is a type as well as constructor, so a simple redirect
to a replacement constructor will not work. Don't try to support it since
it is gone in python 3.

Works with numpy.loadtxt and pandas parsers.

For pandas, either need to specify *engine="python"* or pass an open call
to the reader; if you just pass a filename, then it will try opening it
with the libc open function and fail.  Could potentially monkeypatch
pandas to pre-open the file.
"""

import sys
import os
import os.path
import io
from functools import wraps

try:
    import __builtin__
    from __builtin__ import open as _py2_open
except ImportError:
    __builtin__ = _py2_open = None

try:
    import builtins
except ImportError:
    builtins = None

try:
    from pathlib import PurePath
except ImportError:

    class PurePath:
        pass


# Sphinx hack: in order to avoid having sphinx pick up the docstrings for the
# builtin functions (chdir and listdir don't format properly), simply suppress
# all classes for the module.
__all__ = []

# TODO: restructure according to pathlib interface
# Looking at pathlib, it already provides methods close to what we implement
# for our virtual file system.  By following the pathlib interface directly,
# then any code that is set up to use pathlib can use our virtual file systems
# without change (in particular, without monkeypatching python builtins).

# for functions that work for read-only filesystems, use *fn
# for functions implemented as python, use -fn
#
# os functions
#   *chdir, *getcwd, *listdir, rmdir, mkdir, chroot
#   *open, *stat, *access, rename, link, unlink, remove chmod, chown, chflags
#   *getcwdb, *getcwdu  # bytes getcwd (py3), unicode getcwd (py2)
#   *scandir
#   -makedirs   # uses exists, split, mkdir
#   -removedirs # uses split, rmdir
#   -walk       # uses islink, join, isdir, listdir
#   -glob.glob  # uses lexists, isdir, join, scandir

# symbolic link functions
#   *readlink, *lstat, symlink, lchflags, lchmod, lchown
#   *os.path.lexists

# os.path functions
#   *exists, *isfile, *isdir
#   *get[acm]time, *getsize
#   samefile    # uses stat, samestat
#   samestat    # uses filestat.dev and filestat.inode; compares device/inode
#   -isabs      # uses str; returns s.startswith('/')
#   -normpath   # uses str; pure path manipulation
#   -abspath    # uses isabs, normpath, getcwd/getcwdu
#   -realpath   # uses isabs, split, join, islink, readlink
#   -renames    # uses exists, split, rename, makedirs, removedirs
#   -walk       # uses listdir, lstat;  deprecated in favour of os.walk

# os/os.path constants
#   curdir, pardir, sep, pathsep, defpath, extsep, altsep, linesep

# file descriptor operations in os
#   fchdir, fchmod, fchown, fdopen, close, fstat, fstatvfs, fpathconf,
#   lseek, read, dup, dup2, errno, error, closerange, isatty, openpty,
#   mknod


# Protect against reload(vfs). Assume that if the symbol is already a
# wrapped symbol that we want to retrieve the original unwrapped symbol
# rather than our wrapper. Additionally, assume nobody else is foolish
# enough to be wrapping such low-level symbols...
def _unwrap(fn):
    return getattr(fn, "__wrapped__", fn)


_py2_open = _unwrap(_py2_open)
_py3_open = _unwrap(io.open)
_chdir = _unwrap(os.chdir)
_getcwd = _unwrap(os.getcwd)
_exists = _unwrap(os.path.exists)
_isfile = _unwrap(os.path.isfile)
_isdir = _unwrap(os.path.isdir)
_listdir = _unwrap(os.listdir)
# TODO: maybe use builtin versions?
_abspath = _unwrap(os.path.abspath)
_realpath = _unwrap(os.path.realpath)


class RealFS(object):
    def __enter__(self):
        pushfs(RealFS)

    def __exit__(self, *args, **kw):
        popfs()

    open = _py3_open
    py2_open = _py2_open
    getcwd = _getcwd
    chdir = _chdir
    listdir = _listdir
    abspath = _abspath
    realpath = _realpath
    isfile = _isfile
    isdir = _isdir
    exists = _exists


class ZipFS(object):
    """
    Opens a zip file as the root file system.
    """

    def __init__(self, path):
        import zipfile

        # TODO: can we open a zip within a zip?
        # Apparently yes, but only if we read the file into a byte stream and
        # then work from that file.  See the following stackoverflow answer:
        #    https://stackoverflow.com/questions/12025469/how-to-read-from-a-zip-file-within-zip-file-in-python
        self._path = _realpath(path)
        self._wd = "/"
        self._zip = zipfile.ZipFile(path)

    def __enter__(self):
        pushfs(self)
        return self._zip

    def __exit__(self, *args, **kw):
        popfs()

    def open(self, file, mode="r", buffering=-1, encoding=None, errors=None, newline=None, **kw):
        # abspath handles pathlib
        # Note: python 3 zipfile only supports mode rb; to get unicode
        # decoding, need to wrap the binary stream in a text I/O wrapper.
        zipmode = "r"
        with RealFS():
            fd = self._zip.open(self.abspath(file)[1:], mode=zipmode)
        if "b" in mode:
            return fd
        else:
            return io.TextIOWrapper(fd, encoding=encoding, errors=errors, newline=newline)

    def py2_open(self, name, mode="r", buffering=-1):
        # abspath handles pathlib
        # Note: python 2 zipfile supports modes r, rU, and U, but not rb
        zipmode = "r" if mode == "rb" else mode
        with RealFS():
            fd = self._zip.open(self.abspath(name)[1:], mode=zipmode)
        return fd

    def chdir(self, path):
        # abspath handles pathlib
        if self.isdir(path):
            self._wd = self.abspath(path) + "/"

    def _iter_dir(self, path=None):
        # TODO: turn this into a scandir interface
        # abspath handles pathlib
        path = self._wd if path is None else self.abspath(path)
        path = path[1:]
        n = len(path)
        seen = {}
        for f in self._zip.filelist:
            if not f.filename.startswith(path):
                # it is not part of the tree
                continue
            parts = f.filename.split("/")
            if len(parts) == 1:
                # it is a leaf so report it
                yield parts[0]
            elif parts[0] not in seen:
                # it is a directory, so only report it if it has not already
                # been reported
                seen.add(parts[0])
                yield parts[0]

    def listdir(self, path=None):
        # abspath handles pathlib
        return [f for f in self._iter_dir(path)]

    def abspath(self, path):
        if isinstance(path, PurePath):
            path = path.as_posix()
        if hasattr(path, "decode"):  # CRUFT: python 2
            path = path.decode()
        if path[0] != "/":
            path = "/".join((self._wd[:-1], path))
        return os.path.normpath(path)

    def realpath(self, filename):
        # abspath handles pathlib
        return os.path.join(self._path, self.abspath(filename))

    def isfile(self, path):
        # abspath handles pathlib
        path = self.abspath(path)[1:]
        return any(path == f.filename for f in self._zip.filelist)

    def isdir(self, s):
        # abspath handles pathlib
        path = self.abspath(s)[1:] + "/"
        for f in self._zip.filelist:
            if f.filename.startswith(path):
                return True
        return False

    def exists(self, path):
        # abspath handles pathlib
        return self.isfile(path) or self.isdir(path)


FS = RealFS
FS_STACK = []  # type: List[RealFS]


def pushfs(fs):
    global FS
    FS_STACK.append(FS)
    FS = fs


def popfs():
    global FS
    FS = FS_STACK.pop()


# Note: We need fs redirection functions as bound methods since pathlib uses
# them as class attributes. The usual filesystem methods are all builtin
# functions so they act as static methods when used as part of a class
# definition. However, when python functions are used as class attributes
# they get turned into bound methods when an object is created with self
# as the first parameter. But if the class attribute is already a bound method
# then it is left alone when the object is created.  We still get the extra
# self parameter, but we get it whether we access it as a class attribute
# or as an object attribute so it is easy to ignore.
class VFS(object):
    if _py3_open is not None:

        @wraps(_py3_open)
        def fs_py3_open(self, *args, **kw):
            return FS.open(*args, **kw)

    if _py2_open is not None:

        @wraps(_py2_open)
        def fs_py2_open(self, *args, **kw):
            return FS.py2_open(*args, **kw)

    @wraps(_chdir)
    def fs_chdir(self, *args, **kw):
        return FS.chdir(*args, **kw)

    @wraps(_getcwd)
    def fs_getcwd(self, *args, **kw):
        return FS.getcwd(*args, **kw)

    @wraps(_listdir)
    def fs_listdir(self, *args, **kw):
        return FS.listdir(*args, **kw)

    @wraps(_exists)
    def fs_exists(self, *args, **kw):
        return FS.exists(*args)

    @wraps(_isfile)
    def fs_isfile(self, *args, **kw):
        return FS.isfile(*args, **kw)

    @wraps(_isdir)
    def fs_isdir(self, *args, **kw):
        return FS.isdir(*args, **kw)

    @wraps(_abspath)
    def fs_abspath(self, *args, **kw):
        return FS.abspath(*args, **kw)

    @wraps(_realpath)
    def fs_realpath(self, *args, **kw):
        return FS.realpath(*args, **kw)


vfs = VFS()


def vfs_init():
    """
    Call this very early in your program so that various filesystem functions
    will be redirected even if they are expressed as "from module import fn"
    """
    if __builtin__ is not None:
        __builtin__.open = vfs.fs_py2_open
    if builtins is not None:
        builtins.open = vfs.fs_py3_open
    io.open = vfs.fs_py3_open
    os.chdir = vfs.fs_chdir
    os.getcwd = vfs.fs_getcwd
    os.listdir = vfs.fs_listdir
    os.path.abspath = vfs.fs_abspath
    os.path.realpath = vfs.fs_realpath
    os.path.exists = vfs.fs_exists
    os.path.isfile = vfs.fs_isfile
    os.path.isdir = vfs.fs_isdir

    try:
        import nt, ntpath

        nt.chdir = vfs.fs_chdir
        nt.listdir = vfs.fs_listdir
        nt.getcwd = vfs.fs_getcwd
        ntpath.abspath = vfs.fs_abspath
        ntpath.realpath = vfs.fs_realpath
        ntpath.exists = vfs.fs_exists
        ntpath.isfile = vfs.fs_isfile
        ntpath.isdir = vfs.fs_isdir
    except ImportError:
        pass

    try:
        import posix, posixpath

        posix.chdir = vfs.fs_chdir
        posix.listdir = vfs.fs_listdir
        posix.getcwd = vfs.fs_getcwd
        posixpath.abspath = vfs.fs_abspath
        posixpath.realpath = vfs.fs_realpath
        posixpath.exists = vfs.fs_exists
        posixpath.isfile = vfs.fs_isfile
        posixpath.isdir = vfs.fs_isdir
    except ImportError:
        pass

    # Pathlib may be imported really early.  Make sure it sees the vfs.
    # TODO: with reload some isinstance tests may fail --- monkeypatch instead?
    # Should be just an update to the members of pathlib._NormalAccessor.
    # With reload, os.PathLike.register(PurePath) is being called twice but that
    # shouldn't be a problem, especially because PurePath will have changed.
    try:
        import pathlib
        from importlib import reload

        reload(pathlib)
    except ImportError:
        pass


# CRUFT: use old wrappers for python 2 since new wrappers don't seem to work
if sys.version_info[0] == 2:

    class RealFS(object):
        def __enter__(self):
            pushfs(self)

        def __exit__(self, *args, **kw):
            popfs()

        def open(self, *args, **kw):
            return _py3_open(*args, **kw)

        def py2_open(self, *args, **kw):
            return _py2_open(*args, **kw)

        def getcwd(self):
            return _getcwd()

        def chdir(self, path):
            return _chdir(path)

        def listdir(self, path=None):
            return _listdir(path) if path is not None else _listdir()

        def abspath(self, path):
            return _abspath(path)

        def realpath(self, path):
            return _realpath(path)

        def isfile(self, path):
            return _isfile(path)

        def isdir(self, path):
            return _isdir(path)

        def exists(self, path):
            return _exists(path)

    if _py3_open is not None:

        @wraps(_py3_open)
        def fs_py3_open(*args, **kw):
            return FS.open(*args, **kw)

    if _py2_open is not None:

        @wraps(_py2_open)
        def fs_py2_open(*args, **kw):
            return FS.py2_open(*args, **kw)

    @wraps(_chdir)
    def fs_chdir(*args, **kw):
        return FS.chdir(*args, **kw)

    @wraps(_getcwd)
    def fs_getcwd(*args, **kw):
        return FS.getcwd(*args, **kw)

    @wraps(_listdir)
    def fs_listdir(*args, **kw):
        return FS.listdir(*args, **kw)

    @wraps(_exists)
    def fs_exists(*args, **kw):
        return FS.exists(*args)

    @wraps(_isfile)
    def fs_isfile(*args, **kw):
        return FS.isfile(*args, **kw)

    @wraps(_isdir)
    def fs_isdir(*args, **kw):
        return FS.isdir(*args, **kw)

    @wraps(_abspath)
    def fs_abspath(*args, **kw):
        return FS.abspath(*args, **kw)

    @wraps(_realpath)
    def fs_realpath(*args, **kw):
        return FS.realpath(*args, **kw)

    def vfs_init():
        """
        Call this very early in your program so that various filesystem functions
        will be redirected even if they are expressed as "from module import fn"
        """
        global FS
        FS = RealFS()
        if __builtin__ is not None:
            __builtin__.open = fs_py2_open
        if builtins is not None:
            builtins.open = fs_py3_open
        io.open = fs_py3_open
        os.chdir = fs_chdir
        os.getcwd = fs_getcwd
        os.listdir = fs_listdir
        os.path.abspath = fs_abspath
        os.path.realpath = fs_realpath
        os.path.exists = fs_exists
        os.path.isfile = fs_isfile
        os.path.isdir = fs_isdir

        try:
            import nt, ntpath

            nt.chdir = fs_chdir
            nt.listdir = fs_listdir
            nt.getcwd = fs_getcwd
            ntpath.abspath = fs_abspath
            ntpath.realpath = fs_realpath
            ntpath.exists = fs_exists
            ntpath.isfile = fs_isfile
            ntpath.isdir = fs_isdir
        except ImportError:
            pass

        try:
            import posix, posixpath

            posix.chdir = fs_chdir
            posix.listdir = fs_listdir
            posix.getcwd = fs_getcwd
            posixpath.abspath = fs_abspath
            posixpath.realpath = fs_realpath
            posixpath.exists = fs_exists
            posixpath.isfile = fs_isfile
            posixpath.isdir = fs_isdir
        except ImportError:
            pass

        # Pathlib may be imported really early.  Make sure it sees the vfs.
        # TODO: with reload some isinstance tests may fail --- monkeypatch instead?
        # Should be just an update to the members of pathlib._NormalAccessor.
        # With reload, os.PathLike.register(PurePath) is being called twice but that
        # shouldn't be a problem, especially because PurePath will have changed.
        try:
            import pathlib
            from importlib import reload

            reload(pathlib)
        except ImportError:
            pass
